"""Module to generate overall activity summary from epoch data."""

import pandas as pd
from pandas.tseries.frequencies import to_offset

from actinet.utils.utils import date_parser, to_screen
from actinet.utils.summary_utils import *
from actinet import circadian


def get_activity_summary(
    data,
    labels,
    exclude_daily_wear_below=None,
    intensityDistribution=False,
    circadianMetrics=False,
    verbose=True,
):
    """
    Calculate overall activity summary from predicted activity label data.
    This is achieved by:
    1) calculate imputation values to replace nan PA metric values
    2) calculate empirical cumulative distribution function of vector magnitudes
    3) derive main movement summaries (overall, weekday/weekend, and hour)
    4) derive daily summaries (daily enmo and daily activity)

    :param str data: Input csv.gz file or pandas dataframe of processed epoch data
    :param list(str) labels: Activity state labels
    :param str exclude_daily_wear_below: Exclude days with wear time below this threshold
    :param bool intensityDistribution: Add intensity outputs to dict <summary>
    :param bool circadianMetrics: Add circadian rhythm metrics to dict <summary>
    :param bool verbose: Print verbose output

    :return: A summary of the activity and daily wear statistics.
    :rtype: tuple(dict, pd.DataFrame)

    """
    to_screen("=== Summarizing ===", verbose)

    if isinstance(data, str):
        data = pd.read_csv(
            data,
            index_col=["time"],
            parse_dates=["time"],
            date_parser=date_parser,
        )

    # Impute missing values
    data_imputed = _impute_missing(data, labels, verbose)

    # Main movement summaries
    summary = _summarise(
        data,
        data_imputed,
        labels,
        intensityDistribution,
        circadianMetrics,
        verbose,
    )

    # Daily summaries
    daily_summary = _daily_summary(data, data_imputed, labels, exclude_daily_wear_below, verbose)

    # Return physical activity summaries
    return summary, daily_summary


def _impute_missing(data, labels, verbose=False):
    # In the following, we resample, pad and impute the data so that we have a
    # multiple of 24h for the stats calculations
    to_screen("=== Imputing missing values ===", verbose)
    
    cols = ["acc"] + labels
    if "MET" in data.columns:
        cols.append("MET")
    data_imputed = impute_missing(data[cols].astype("float"))

    return data_imputed


def _summarise(
    data,
    data_imputed,
    labels,
    intensityDistribution=False,
    circadianMetrics=False,
    verbose=False,
    summary={},
):
    """Overall summary stats for each activity type to summary dict

    :param pandas.DataFrame data: Pandas dataframe of epoch data
    :param pandas.DataFrame data_adjusted: Pandas dataframe of epoch data with imputed missing values
    :param list(str) labels: Activity state labels
    :param dict summary: Output dictionary containing all summary metrics
    :param bool intensityDistribution: Add intensity outputs to dict <summary>
    :param bool circadianMetrics: Add circadian rhythm metrics to dict <summary>
    :param bool verbose: Print verbose output
    :param dict summary: Output dictionary containing all summary metrics

    :return: Updated dict <summary> keys for each activity type 'overall-<avg/sd>',
        'week<day/end>-avg', '<day..>-avg', 'hourOfDay-<hr..>-avg',
        'hourOfWeek<day/end>-<hr..>-avg'
    :rtype: dict
    """

    data = data.copy()
    freq = to_offset(infer_freq(data.index))

    # Get start day
    startTime = data.index[0]
    summary["FirstDay(0=mon,6=sun)"] = startTime.weekday()

    # Hours of activity for each recorded day
    epochPeriod = int(pd.Timedelta(freq).total_seconds())
    cols = labels
    dailyStats = (
        data[cols].astype("float").groupby(data.index.date).sum() * epochPeriod / 3600
    ).reset_index(drop=True)

    for i, row in dailyStats.iterrows():
        for col in cols:
            summary[f"day{i}-recorded-{col}(hrs)"] = row.loc[col]

    # Calculate empirical cumulative distribution function of vector magnitudes
    if intensityDistribution:
        summary = calculateECDF(data["acc"], summary)

    tStart, tEnd = data.index[0], data.index[-1]

    # Overall stats (no padding, i.e. only within recording period)
    to_screen("=== Calculating overall statistics ===", verbose)
    overallStats = data_imputed[tStart:tEnd].apply(["mean", "std"])
    for col in overallStats:
        summary[f"{col}-overall-avg"] = overallStats[col].loc["mean"]
        summary[f"{col}-overall-sd"] = overallStats[col].loc["std"]

    dayOfWeekStats = data_imputed.groupby([data_imputed.index.weekday, 
                                           data_imputed.index.hour]).mean()
    dayOfWeekStats.index = dayOfWeekStats.index.set_levels(
        dayOfWeekStats.index.levels[0]
        .to_series()
        .replace({0: "mon", 1: "tue", 2: "wed", 3: "thu", 4: "fri", 5: "sat", 6: "sun"})
        .to_list(),
        level=0,
    )
    dayOfWeekStats.index.set_names(["DayOfWeek", "Hour"], inplace=True)

    # Week stats
    for col, value in dayOfWeekStats.mean().items():
        summary[f"{col}-week-avg"] = value

    # Stats by day of week (Mon, Tue, ...)
    for col, stats in dayOfWeekStats.groupby(level=0).mean().to_dict().items():
        for dayOfWeek, value in stats.items():
            summary[f"{col}-{dayOfWeek}-avg"] = value

    # Stats by hour of day
    for col, stats in dayOfWeekStats.groupby(level=1).mean().to_dict().items():
        for hour, value in stats.items():
            summary[f"{col}-hourOfDay-{hour}-avg"] = value

    weekdayOrWeekendStats = dayOfWeekStats.groupby(
        [
            dayOfWeekStats.index.get_level_values("DayOfWeek").str.contains("sat|sun"),
            dayOfWeekStats.index.get_level_values("Hour"),
        ]
    ).mean()
    weekdayOrWeekendStats.index = weekdayOrWeekendStats.index.set_levels(
        weekdayOrWeekendStats.index.levels[0]
        .to_series()
        .replace({True: "Weekend", False: "Weekday"})
        .to_list(),
        level=0,
    )
    weekdayOrWeekendStats.index.set_names(["WeekdayOrWeekend", "Hour"], inplace=True)

    # Weekday/weekend stats
    for col, stats in weekdayOrWeekendStats.groupby(level=0).mean().to_dict().items():
        for weekdayOrWeekend, value in stats.items():
            summary[f"{col}-{weekdayOrWeekend.lower()}-avg"] = value

    # Stats by hour of day AND by weekday/weekend
    for col, stats in weekdayOrWeekendStats.to_dict().items():
        for key, value in stats.items():
            weekdayOrWeekend, hour = key
            summary[f"{col}-hourOf{weekdayOrWeekend}-{hour}-avg"] = value

    # Calculate circadian metrics
    if circadianMetrics:
        to_screen("=== Calculating circadian metrics ===", verbose)
        summary = circadian.calculatePSD(data_imputed, epochPeriod, False, labels, summary)
        summary = circadian.calculatePSD(data_imputed, epochPeriod, True, labels, summary)
        summary = circadian.calculateFourierFreq(
            data_imputed, epochPeriod, False, labels, summary
        )
        summary = circadian.calculateFourierFreq(
            data_imputed, epochPeriod, True, labels, summary
        )
        summary = circadian.calculateM10L5(data_imputed, epochPeriod, summary)

    return summary


def _daily_summary(data, data_imputed, labels, exclude_daily_wear_below, verbose=False):
    to_screen("=== Daily summary ===", verbose)

    min_wear_per_day = 0 if exclude_daily_wear_below is None \
        else pd.Timedelta(exclude_daily_wear_below).total_seconds() / 60 # in minutes
    
    daily_enmo = summarize_daily_enmo(data["acc"], data_imputed["acc"], min_wear_per_day)
    daily_activity = summarize_daily_activity(data[labels], data_imputed[labels], labels, min_wear_per_day)
    daily_summary = pd.concat([daily_enmo, daily_activity], axis=1)
    daily_summary.index.name = "Date"

    return daily_summary



import numpy as np
import pandas as pd

SLEEP_GAP_TOLERANCE = 30 * 60  # 30 minutes
SLEEP_BLOCK_PERIOD = 24 * 60 * 60  # 24 hours

HMM_SEDENTARY_CODE = 2
HMM_SLEEP_CODE = 3


def removeSpuriousSleep(Y, labels, period, sleepTol='1H', removeNaps=False):
    """
    Remove spurious sleep epochs from activity classification.

    :param numpy.ndarray Y: Sequence of activity labels
    :param list labels: List of activity labels
    :param int period: Time period between successive labels in seconds
    :param str sleepTol: Minimum sleep duration, e.g. '1H'
    :param bool removeNaps: If True, restricts sleep to the longest sleep block each 24 hours.
    :return: Numpy array of revised model output
    :rtype: numpy.ndarray

    """
    label_map = {label: i for i, label in enumerate(labels)}

    if not sleepTol and not removeNaps:
        return Y

    try:
        sleep_code = label_map['sleep']
        sedentary_code = label_map['sedentary'] if 'sedentary' in labels else label_map['sit-stand']
    except KeyError:
        raise ValueError(f"'sleep' and 'sedentary' or 'sit-stand' must be output labels for spurious sleep correction.")

    if sleepTol:
        Y = convertSleepBelowThreshold(Y, period, sleep_code, sedentary_code, sleepTol)

    if removeNaps:
        Y = convertNaps(Y, period, sleep_code, sedentary_code)

    return Y


def convertSleepBelowThreshold(Y, period, sleep_code, sedentary_code, sleepTol='1H'):
    """
    Convert sleep labels to sedentary if sleep block is below an expected threshold.

    :param numpy.ndarray Y: Sequence of activity labels
    :param int period: Time period between successive labels in seconds
    :param int sleep_code: Code for sleep in the labels
    :param int sedentary_code: Code for sedentary in the labels
    :param str sleepTol: Minimum sleep duration, e.g. '1H'

    :return: Numpy array of revised model output
    :rtype: numpy.ndarray

    """
    Y_series = pd.Series(Y.copy())
    sleep_mask = Y_series == sleep_code

    if sleep_mask.sum() == 0:
        return Y

    sleepStreak = (
        sleep_mask.ne(sleep_mask.shift())
        .cumsum()
        .pipe(lambda x: x.groupby(x).transform('count') * sleep_mask)
    )

    short_sleep = sleep_mask & (sleepStreak < (pd.Timedelta(sleepTol).total_seconds() / period))
    Y_series.loc[short_sleep] = sedentary_code

    return Y_series.values


def convertNaps(Y, period, sleep_code, sedentary_code):
    """
    Convert sleep labels to sedentary, if they do not occur during the longest sleep block in each 24-hour period.

    :param numpy.ndarray Y: Sequence of activity labels
    :param int period: Time between labels in seconds
    :param int sleep_code: Label code indicating sleep
    :param int sedentary_code: Label code indicating sedentary

    :return: Revised numpy array with naps removed
    :rtype: numpy.ndarray
    """

    sleep_blocks = find_blocks(Y, gap_tol=int(SLEEP_GAP_TOLERANCE/period),
                               block_code=sleep_code)

    longest_blocks = select_longest_blocks_per_period(
        sleep_blocks, len(Y), block_period=int(SLEEP_BLOCK_PERIOD/period)
    )

    return convert_non_selected_block(Y, longest_blocks, sleep_code, sedentary_code)


def find_blocks(labels, gap_tol, block_code='s'):
    """Finds blocks of a specific code in a sequence of labels, allowing for gap tolerance of other labels."""
    blocks = []
    is_block = False
    gap_len = 0

    for i, elem in enumerate(labels):
        if not is_block and elem == block_code:
            is_block = True
            block_start = i
            gap_len = 0

        elif is_block:
            if elem == block_code:
                gap_len = 0

            else:
                gap_len += 1
                if gap_len == gap_tol:
                    blocks.append((block_start, i - gap_tol))
                    is_block = False

        if is_block and i == len(labels) - 1:
            blocks.append((block_start, i))

    return blocks


def extract_start_end_tuple(row):
    """Helper to extract a (start, end) tuple from a DataFrame row."""
    return (row['start'], row['end'])


def select_longest_blocks_per_period(blocks, sequence_len, block_period):
    """
    Selects the longest blocks within an expected period.
    Also includes edge blocks that start at 0 or end at sequence end.
    """
    
    if not blocks:
        return []

    blocks_df = pd.DataFrame(blocks, columns=['start', 'end'])
    blocks_df['length'] = blocks_df['end'] - blocks_df['start'] + 1

    selected = []
    ref_start = 0

    if (start_blocks := blocks_df[blocks_df['start'] == 0]).any().any():
        selected.append(extract_start_end_tuple(start_blocks.iloc[0]))
        ref_start = start_blocks.iloc[0]['end']
    if (end_blocks := blocks_df[blocks_df['end'] == sequence_len - 1]).any().any():
        selected.append(extract_start_end_tuple(end_blocks.iloc[0]))

    blocks_df = blocks_df[(blocks_df['start'] != 0) & (blocks_df['end'] != sequence_len - 1)]

    while ref_start + block_period < sequence_len:
        window_blocks = blocks_df[(blocks_df['start'] > ref_start) & (blocks_df['start'] <= ref_start + block_period)]
        if window_blocks.empty:
            ref_start += block_period
            continue
        
        best_block = window_blocks.loc[window_blocks['length'].idxmax()]
        
        selected.append(extract_start_end_tuple(best_block))
        ref_start = best_block['end']

    return selected


def convert_non_selected_block(labels, selected_blocks, block_code='s', conv_code='d'):
    """
    Converts all block code labels found outside longest blocks to conv code.
    """
    mask = np.zeros(len(labels), dtype=bool)

    for start, end in selected_blocks:
        mask[start:end+1] = True

    mask_to_change = (labels == block_code) & (~mask)
    labels[mask_to_change] = conv_code

    return labels


def add_sleep_sedentary_transitions(df):
    """ 
    Adds a single transition from sleep to sedentary, and vice-versa, for each participant, if it does not exist.
    """
    df_copy = df.copy()

    rows_to_append = []

    for group in df_copy["group"].unique():
        group_df = df_copy[df_copy["group"] == group]

        if not ((group_df["label"] == HMM_SLEEP_CODE) & (group_df["shift"] == HMM_SEDENTARY_CODE)).any():
            rows_to_append.append({"label": HMM_SLEEP_CODE, "shift": HMM_SEDENTARY_CODE, "group": group})

        if not ((group_df["label"] == HMM_SEDENTARY_CODE) & (group_df["shift"] == HMM_SLEEP_CODE)).any():
            rows_to_append.append({"label": HMM_SEDENTARY_CODE, "shift": HMM_SLEEP_CODE, "group": group})

    if rows_to_append:
        new_rows = pd.DataFrame(rows_to_append)
        df = pd.concat([df, new_rows], ignore_index=True)

    return df



import datetime
import re
import numpy as np
import pandas as pd
from scipy.interpolate import interp1d
from typing import Union


ACTIVITY_LABELS_DICT = {
    'Walmsley2020': {
        'sleep': 'Sleep',
        'sedentary': 'Sedentary Behaviour',
        'light': 'Light Activity',
        'moderate-vigorous': 'Moderate-Vigorous Activity'
    },
    'Willetts2018': {
        'sleep': 'Sleep',
        'sit-stand': 'Sit/Stand',
        'vehicle': 'Vehicle',
        'walking': 'Walking',
        'mixed': 'Mixed Activity',
        'bicycling': 'Bicycling'
    }
}


def infer_freq(t):
    """Like pd.infer_freq but more forgiving"""
    tdiff = t.to_series().diff()
    q1, q3 = tdiff.quantile([0.25, 0.75])
    tdiff = tdiff[(q1 <= tdiff) & (tdiff <= q3)]
    freq = tdiff.mean()
    freq = pd.Timedelta(freq)
    return freq


def to_screen(msg, verbose=True):
    """
    Print msg str prepended with current time

    :param str mgs: Message to be printed to screen
    :return: None. Prints msg str prepended with current time.
    """
    if verbose:
        timeFormat = "%Y-%m-%d %H:%M:%S"
        print(f"\n{datetime.datetime.now().strftime(timeFormat)}\t{msg}")


def date_parser(t):
    """
    Parse date a date string of the form e.g.
    2020-06-14 19:01:15.123+0100 [Europe/London]
    """
    tz = re.search(r"(?<=\[).+?(?=\])", t)
    if tz is not None:
        tz = tz.group()
    t = re.sub(r"\[(.*?)\]", "", t)
    return pd.to_datetime(t, utc=True).tz_convert(tz)


def safe_indexer(array, indexes):
    return array[indexes] if array is not None else None


def is_good_window(x, window_len, columns):
    """
    Check if a window is considered good based on its length, the presence of NaN values in specified columns.

    Args:
        x (pd.DataFrame): Window data.
        window_len (int): The index length of the data.
        columns (list): List of relevant columns to check for NaN values.

    Returns:
        bool: True if the window is considered good, False otherwise.

    """
    if not all(col in x.columns for col in columns):
        return False

    # Check window length is correct
    if len(x[columns]) != window_len:
        return False

    # Check no nans
    if pd.isna(x[columns]).any(axis=None):

        return False

    return True


def resize(x, length, axis=1):
    """
    Resize the temporal length of the data using linear interpolation.

    Args:
        x (ndarray): Data to be resized.
        length (int): New length of the data.
        axis (int): Axis along which to perform the interpolation. Defaults to 1.

    Returns:
        ndarray: Resized data.

    """
    length_orig = x.shape[axis]
    t_orig = np.linspace(0, 1, length_orig, endpoint=True)
    t_new = np.linspace(0, 1, length, endpoint=True)
    x = interp1d(t_orig, x, kind="linear", axis=axis, assume_sorted=True)(t_new)
    return x


def drop_first_last_days(
    x: Union[pd.Series, pd.DataFrame],
    first_or_last='both'
):
    """
    Drop the first day, last day, or both from a time series.

    Parameters:
    - x (pd.Series or pd.DataFrame): A pandas Series or DataFrame with a DatetimeIndex representing time series data.
    - first_or_last (str, optional): A string indicating which days to drop. Options are 'first', 'last', or 'both'. Default is 'both'.

    Returns:
    - pd.Series or pd.DataFrame: A pandas Series or DataFrame with the values of the specified days dropped.

    Example:
        # Drop the first day from the series
        series = drop_first_last_days(series, first_or_last='first')
    """
    if len(x) == 0:
        print("No data to drop")
        return x

    if first_or_last == 'first':
        x = x[x.index.date != x.index.date[0]]
    elif first_or_last == 'last':
        x = x[x.index.date != x.index.date[-1]]
    elif first_or_last == 'both':
        x = x[(x.index.date != x.index.date[0]) & (x.index.date != x.index.date[-1])]
    return x


def flag_wear_below_days(
    x: Union[pd.Series, pd.DataFrame],
    min_wear: str = '12H'
):
    """
    Set days containing less than the specified minimum wear time (`min_wear`) to NaN.

    Parameters:
    - x (pd.Series or pd.DataFrame): A pandas Series or DataFrame with a DatetimeIndex representing time series data.
    - min_wear (str): A string representing the minimum wear time required per day (e.g., '8H' for 8 hours).

    Returns:
    - pd.Series or pd.DataFrame: A pandas Series or DataFrame with days having less than `min_wear` of valid data set to NaN.

    Example:
        # Exclude days with less than 12 hours of valid data
        series = exclude_wear_below_days(series, min_wear='12H')
    """
    if len(x) == 0:
        print("No data to exclude")
        return x

    min_wear = pd.Timedelta(min_wear)
    dt = infer_freq(x.index)
    ok = x.notna()
    if isinstance(ok, pd.DataFrame):
        ok = ok.all(axis=1)
    ok = (
        ok
        .groupby(x.index.date)
        .sum() * dt
        >= min_wear
    )
    # keep ok days, rest is set to NaN
    x = x.copy()  # make a copy to avoid modifying the original data
    x[np.isin(x.index.date, ok[~ok].index)] = np.nan
    return x


def calculate_wear_stats(data: pd.DataFrame):
    """
    Calculate wear time and related information from raw accelerometer data.

    Parameters:
    - data (pd.DataFrame): A pandas DataFrame of raw accelerometer data with columns 'x', 'y', 'z' and a DatetimeIndex.

    Returns:
    - dict: A dictionary containing various wear time stats.

    Example:
        info = calculate_wear_stats(data)
    """

    TIME_FORMAT = "%Y-%m-%d %H:%M:%S"

    n_data = len(data)

    if n_data == 0:
        start_time = None
        end_time = None
        wear_start_time = None
        wear_end_time = None
        nonwear_duration = 0.0
        wear_duration = 0.0
        covers24hok = 0

    else:
        na = data.isna().any(axis=1)  # TODO: check na only on x,y,z cols?
        dt = infer_freq(data.index).total_seconds()
        start_time = data.index[0].strftime(TIME_FORMAT)
        end_time = data.index[-1].strftime(TIME_FORMAT)
        wear_start_time = data.first_valid_index()
        if wear_start_time is not None:
            wear_start_time = wear_start_time.strftime(TIME_FORMAT)
        wear_end_time = data.last_valid_index()
        if wear_end_time is not None:
            wear_end_time = wear_end_time.strftime(TIME_FORMAT)
        nonwear_duration = na.sum() * dt / (60 * 60 * 24)
        wear_duration = n_data * dt / (60 * 60 * 24) - nonwear_duration 
        coverage = (~na).groupby(na.index.hour).mean()
        covers24hok = int(len(coverage) == 24 and coverage.min() >= 0.01)

    return {
        'StartTime': start_time,
        'EndTime': end_time,
        'WearStartTime': wear_start_time,
        'WearEndTime': wear_end_time,
        'WearTime(days)': wear_duration,
        'NonwearTime(days)': nonwear_duration,
        'Covers24hOK': covers24hok
    }



import argparse
import os


def generate_commands(
    input_dir, output_dir, cmdsfile="list-of-commands.txt", fext="cwa", cmdopts=""
):
    """Generate a text file listing processing commands for files found under input_dir/

    :param str input_dir: Directory containing accelerometer files to process.
    :param str output_dir: Name for output directory to store the processing results.
    :param str cmdsfile: Name for generated file with list of commands.
    :param str fext: Accelerometer file extension e.g. cwa, CWA, bin, BIN, gt3x...
    :param str cmdopts: String of options to pass e.g. "--type rf" to use Random Forest classifier.

    :return: New file written to <cmdsfile>
    :rtype: void

    :Example:
    >>> import utils
    >>> utils.generate_commands("MyAccFiles/", output_dir="MyOutputs/", cmdsfile="list-of-commands.txt", fext="cwa", cmdopts="--type rf")
    <list of commands written to "list-of-commands.txt">
    """

    # List all accelerometer files under input_dir/
    fpaths = []
    fext = fext.lower()
    for root, _, files in os.walk(input_dir):
        for file in files:
            if file.lower().endswith(
                (fext, fext + ".gz", fext + ".zip", fext + ".bz2", fext + ".xz")
            ):
                fpaths.append(os.path.join(root, file))
    print(f"Found {len(fpaths)} accelerometer files...")

    with open(cmdsfile, "w") as f:
        for fpath in fpaths:

            # Use the file name as the output folder name for the process,
            # keeping the same directory structure of input_dir/
            # Example: If fpath is {input_dir}/group0/subject123.cwa then
            # _output_dir will be {output_dir}/group0/subject123/
            _output_dir = fpath.replace(
                input_dir.rstrip("/"), output_dir.rstrip("/")
            ).split(".")[0]

            cmd = f"stepcount '{fpath}' --outdir '{_output_dir}' {cmdopts}"

            f.write(cmd)
            f.write("\n")

    print("List of commands written to ", cmdsfile)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("input_dir")
    parser.add_argument("--output_dir", "-d", required=True)
    parser.add_argument("--cmdsfile", "-f", type=str, default="list-of-commands.txt")
    parser.add_argument(
        "--fext",
        "-a",
        default="cwa",
        help="Acc file type e.g. cwa, CWA, bin, BIN, gt3x...",
    )
    parser.add_argument(
        "--cmdopts",
        "-x",
        type=str,
        default="",
        help="String of processing options e.g. '--type rf' to use Random Forest classifier.",
    )
    args = parser.parse_args()

    generate_commands(
        input_dir=args.input_dir,
        output_dir=args.output_dir,
        cmdsfile=args.cmdsfile,
        fext=args.fext,
        cmdopts=args.cmdopts,
    )


if __name__ == "__main__":
    main()



import pandas as pd
from pandas.tseries.frequencies import to_offset
import numpy as np
from actinet.utils.utils import infer_freq


def impute_missing(data, extrapolate=True):
    """Impute missing/nonwear segments

    Impute non-wear data segments using the average of similar time-of-day values
    with one minute granularity on different days of the measurement. This
    imputation accounts for potential wear time diurnal bias where, for example,
    if the device was systematically less worn during sleep in an individual,
    the crude average vector magnitude during wear time would be a biased
    overestimate of the true average. See
    https://journals.plos.org/plosone/article?id=10.1371/journal.pone.0169649#sec013

    :param pandas.DataFrame e: Pandas dataframe of epoch data
    :param bool verbose: Print verbose output

    :return: Update DataFrame <e> columns nan values with time-of-day imputation
    :rtype: void
    """

    if extrapolate:
        # padding at the boundaries to have full 24h
        data = data.reindex(
            pd.date_range(
                data.index[0].floor("D"),
                data.index[-1].ceil("D"),
                freq=to_offset(pd.infer_freq(data.index)),
                inclusive="left",
                name="time",
            ),
            method="nearest",
            tolerance=pd.Timedelta("1m"),
            limit=1,
        )

    def fillna(subframe):
        # Transform will first pass the subframe column-by-column as a Series.
        # After passing all columns, it will pass the entire subframe again as a DataFrame.
        # Processing the entire subframe is optional (return value can be omitted). See 'Notes' in transform doc.
        if isinstance(subframe, pd.Series):
            x = subframe.to_numpy()
            nan = np.isnan(x)
            nanlen = len(x[nan])
            if 0 < nanlen < len(x):  # check x contains a NaN and is not all NaN
                x[nan] = np.nanmean(x)
                return x  # will be cast back to a Series automatically
            else:
                return subframe

    data = (
        data
        # first attempt imputation using same day of week
        .groupby([data.index.weekday, data.index.hour, data.index.minute])
        .transform(fillna)
        # then try within weekday/weekend
        .groupby([data.index.weekday >= 5, data.index.hour, data.index.minute])
        .transform(fillna)
        # finally, use all other days
        .groupby([data.index.hour, data.index.minute])
        .transform(fillna)
    )

    return data


def calculateECDF(x, summary):
    """Calculate activity intensity empirical cumulative distribution

    The input data must not be imputed, as ECDF requires different imputation
    where nan/non-wear data segments are IMPUTED FOR EACH INTENSITY LEVEL. Here,
    the average of similar time-of-day values is imputed with one minute
    granularity on different days of the measurement. Following intensity levels
    are calculated:
    1mg bins from 1-20mg
    5mg bins from 25-100mg
    25mg bins from 125-500mg
    100mg bins from 500-2000mg

    :param pandas.DataFrame e: Pandas dataframe of epoch data
    :param str inputCol: Column to calculate intensity distribution on
    :param dict summary: Output dictionary containing all summary metrics

    :return: Updated summary file
    :rtype: dict
    """

    levels = np.concatenate(
        [
            np.linspace(1, 20, 20),  # 1mg bins from 1-20mg
            np.linspace(25, 100, 16),  # 5mg bins from 25-100mg
            np.linspace(125, 500, 16),  # 25mg bins from 125-500mg
            np.linspace(600, 2000, 15),  # 100mg bins from 500-2000mg
        ]
    ).astype("int")

    whrnan = x.isna().to_numpy()
    ecdf = x.to_numpy().reshape(-1, 1) <= levels.reshape(1, -1)
    ecdf[whrnan] = np.nan

    ecdf = (
        pd.DataFrame(ecdf, index=x.index, columns=levels)
        .groupby([x.index.hour, x.index.minute])
        .mean()  # first average is across same time of day
        .mean()  # second average is within each level
    )

    # Write to summary
    for level, val in ecdf.items():
        summary[f"{x.name}-ecdf-{level}mg"] = val

    return summary


def summarize_daily_enmo(
    acc: pd.Series,
    acc_adjusted: pd.Series,
    min_wear_per_day: float = 21 * 60
):
    """
    Summarize daily ENMO information from raw accelerometer data.

    Parameters:
    - data (pd.Series): A pandas Series of ENMO values for each window.
    - data_adjusted (pd.Series): A pandas Series of ENMO values for each window, with missingness imputed.
    - min_wear_per_day (float, optional): The minimum required wear time (in minutes) for a day to be considered valid. Defaults to 21 hours.

    Returns:
    - pd.DataFrame: A DataFrame with dates as index and daily wear statistics as columns:
        - 'ENMO(mg)': Total ENMO in mg
        - 'ENMO Adjusted(mg)': Total ENMO in mg, adjusted for non-wear

    Example:
        summary = summarize_daily_enmo(acc, acc_adjusted, min_wear_per_day=21*60)
    """

    def _is_enough(x, min_wear=None, dt=None):
        if min_wear is None:
            return True  # no minimum wear time, then default to True
        if dt is None:
            dt = infer_freq(x.index).total_seconds()
        return x.notna().sum() * dt / 60 > min_wear

    def _mean(x, min_wear=None, dt=None):
        if not _is_enough(x, min_wear, dt):
            return np.nan
        return x.mean()

    dt = infer_freq(acc.index).total_seconds()

    if infer_freq(acc_adjusted.index) != infer_freq(acc.index):
        raise ValueError("Data and Data_adjusted must have the same frequency")

    daily = acc.resample('D').agg(_mean, min_wear=min_wear_per_day, dt=dt).rename('ENMO(mg)')
    daily_adj = acc_adjusted.resample('D').agg(_mean, min_wear=min_wear_per_day, dt=dt).rename('ENMO Adjusted(mg)')
    
    summary = pd.concat([daily, daily_adj], axis=1)

    return summary


def summarize_daily_activity(data: pd.DataFrame, data_adjusted: pd.DataFrame, 
                             labels: list, min_wear_per_day: float = 21 * 60):
    """
    Summarize daily activity information from predicted label outputs.

    Parameters:
    - data (pd.DataFrame): A pandas DataFrame of predicted activity labels with a DatetimeIndex.
    - data_adjusted (pd.DataFrame): A pandas DataFrame of predicted activity labels with a DatetimeIndex, with missingness imputed.
    - labels (list): List of activity state labels (e.g., ['sleep', 'sedentary', 'light', 'moderate-vigorous']).
    - min_wear_per_day (float, optional): The minimum required wear time (in minutes) for a day to be considered valid. Defaults to 21 hours.

    Returns:
    - pd.DataFrame: A DataFrame with dates as index and daily activity statistics as columns:
        - 'Sleep (hours)': Total sleep duration in hours
        - 'Sedentary (hours)': Sedentary activity duration in hours
        - 'Light (hours)': Light activity duration in hours
        - 'MVPA (hours)': Moderate activity duration in hours
        - 'Sleep Adjusted (hours)': Total sleep duration in hours, adjusted for non-wear
        - 'Sedentary Adjusted (hours)': Sedentary activity duration in hours, adjusted for non-wear
        - 'Light Adjusted (hours)': Light activity duration in hours, adjusted for non-wear
        - 'MVPA Adjusted (hours)': Moderate activity duration in hours, adjusted for non-wear
    """
    if infer_freq(data_adjusted.index) != infer_freq(data.index):
        raise ValueError("Data and Data_adjusted must have the same frequency")
    
    dt = infer_freq(data.index).total_seconds()
    
    def _is_enough(x, min_wear=None, dt=None):
        if min_wear is None:
            return True  # no minimum wear time, then default to True
        if dt is None:
            dt = infer_freq(x.index).total_seconds()
        return x.notna().sum() * dt / 60 > min_wear

    def _total_hrs(x, min_wear=None, dt=None):
        if not _is_enough(x, min_wear, dt):
            return np.nan
        return x.sum() * dt / 3600

    daily = data.resample('D')\
                .agg(_total_hrs, min_wear=min_wear_per_day, dt=dt)\
                .rename(columns={label: f"{label.capitalize()}(hours)" for label in labels})
    daily_adj = data_adjusted.resample('D')\
                             .agg(_total_hrs, min_wear=min_wear_per_day, dt=dt)\
                             .rename(columns={label: f"{label.capitalize()} Adjusted(hours)"
                                                for label in labels})

    summary = pd.concat([daily, daily_adj], axis=1)

    return summary


def calculate_daily_wear_stats(data: pd.DataFrame):
    """
    Calculate daily wear time statistics from raw accelerometer data.

    Parameters:
    - data (pd.DataFrame): A pandas DataFrame of raw accelerometer data with columns 'x', 'y', 'z' and a DatetimeIndex.

    Returns:
    - pd.DataFrame: A DataFrame with dates as index and daily wear statistics as columns:
        - 'WearTime(hours)': Total wear time in hours

    Example:
        daily_wear_stats = calculate_daily_wear_stats(data)
    """

    if len(data) == 0:
        return pd.DataFrame()

    # Identify non-wear periods (NaN in any of x,y,z columns)
    na = data.isna().any(axis=1)
    dt = infer_freq(data.index).total_seconds()

    # Group by date
    date_groups = data.groupby(data.index.date)

    results = []

    for date, day_data in date_groups:
        day_na = na.loc[day_data.index]
        n_samples = len(day_data)

        if n_samples == 0:
            # Skip empty days
            continue

        # Calculate wear time
        nonwear_samples = day_na.sum()
        wear_samples = n_samples - nonwear_samples

        # Convert to hours
        wear_hours = wear_samples * dt / 3600

        results.append({
            'Date': pd.to_datetime(date),
            'WearTime(hours)': round(wear_hours, 2)
        })

    if not results:
        return pd.DataFrame()

    # Create DataFrame and set date as index
    daily_stats = pd.DataFrame(results)
    daily_stats.set_index('Date', inplace=True)
    daily_stats.index.name = 'Date'

    return daily_stats


import numpy as np
import os
import pandas as pd

from actinet.utils.sleep_utils import add_sleep_sedentary_transitions


class HMM:
    """
    Implement a basic hidden Markov model (HMM) with parameter saving/loading
    https://en.wikipedia.org/wiki/Hidden_Markov_model.
    """

    def __init__(
        self,
        prior=None,
        emission=None,
        transition=None,
        labels=None,
        uniform_prior=True,
        ignore_transition_gaps=False,
        handle_sleep_transitions=False
    ):
        self.prior = prior
        self.emission = emission
        self.transition = transition
        self.labels = labels
        self.uniform_prior = uniform_prior
        self.ignore_transition_gaps = ignore_transition_gaps
        self.handle_sleep_transitions = handle_sleep_transitions

    def __str__(self):
        return (
            "Hidden Markov Model\n"
            "Ignore transition gaps: {self.ignore_transition_gaps}\n"
            "Correct sleep transitions: {self.handle_sleep_transitions}\n"
            "prior: {self.prior}\n"
            "emission: {self.emission}\n"
            "transition: {self.transition}".format(self=self)
        )

    def fit(self, Y_prob, Y_true, groups=None, T=None, interval=None):
        """
        Fit a HMM to the provided data by calculating the prior, transition and emission matrices.

        :param Y_prob: Observation probabilities
        :type Y_prob: numpy.ndarray
        :param Y_true: Ground truth labels
        :type Y_true: numpy.ndarray
        :param groups: Group labels for the data, if applicable
        :type groups: numpy.ndarray, optional
        :param T: Time at each observation
        :type T: numpy.ndarray, optional
        :param interval: Expected time interval between observations in seconds
        :type interval: int or float, optional
        """

        if self.labels is None:
            self.labels = np.unique(Y_true)

        prior = np.mean(Y_true.reshape(-1, 1) == self.labels, axis=0)

        emission = np.vstack(
            [np.mean(Y_prob[Y_true == label], axis=0) for label in self.labels]
        )

        transition = calculate_transition_matrix(Y_true, groups, T, interval, self.ignore_transition_gaps, 
                                                 self.handle_sleep_transitions)

        self.prior = prior
        self.emission = emission
        self.transition = transition

    def predict(self, y_obs, t=None, interval=None, uniform_prior=None):
        """
        Predict sequence of activities using viterbi algorithm, while restoring labels after gaps in data.

        :param y_obs: Predicted observations
        :type y_obs: numpy.ndarray
        :param t: Time at each observation
        :type t: numpy.ndarray, optional
        :param interval: Expected time interval between observations in seconds
        :type interval: int or float, optional
        :param uniform_prior: Assume uniform priors.
        :type uniform_prior: bool, optional

        :return: Smoothed sequence of activities
        :rtype: np.ndarray
        """
        check_for_input_errors(y_obs, t, interval,
                               ignore_transition_gaps=self.ignore_transition_gaps,
                               handle_sleep_transitions=False)

        y_smooth = self.viterbi(y_obs, uniform_prior)

        if not self.ignore_transition_gaps:
            y_smooth = restore_labels_after_gaps(y_obs, y_smooth, t, interval)

        return y_smooth

    def viterbi(self, y_obs, uniform_prior=None):
        """Perform HMM smoothing over observations via Viteri algorithm
        https://en.wikipedia.org/wiki/Viterbi_algorithm.

        :param y_obs: Predicted observations
        :type y_obs: numpy.ndarray
        :param uniform_prior: Assume uniform priors
        :type uniform_prior: bool, optional

        :return: Smoothed sequence of activities
        :rtype: np.ndarray
        """

        def log(x):
            return np.log(x + 1e-16)

        prior = (
            np.ones(len(self.labels)) / len(self.labels)
            if (self.uniform_prior or uniform_prior)
            else self.prior
        )
        emission = self.emission
        transition = self.transition
        labels = self.labels

        nobs = len(y_obs)
        n_labels = len(labels)

        y_obs = np.where(y_obs.reshape(-1, 1) == labels)[1]  # to numeric

        probs = np.zeros((nobs, n_labels))
        probs[0, :] = log(prior) + log(emission[:, y_obs[0]])
        for j in range(1, nobs):
            for i in range(n_labels):
                probs[j, i] = np.max(
                    log(emission[i, y_obs[j]]) + log(transition[:, i]) + probs[j - 1, :]
                )  # probs already in log scale
        viterbi_path = np.zeros_like(y_obs)
        viterbi_path[-1] = np.argmax(probs[-1, :])
        for j in reversed(range(nobs - 1)):
            viterbi_path[j] = np.argmax(
                log(transition[:, viterbi_path[j + 1]]) + probs[j, :]
            )  # probs already in log scale

        viterbi_path = labels[viterbi_path]  # to labels

        return viterbi_path

    def save(self, path):
        """
        Save model parameters to a Numpy npz file.

        :param path: npz file location
        :type path: str
        """
        os.makedirs(os.path.dirname(path), exist_ok=True)
        np.savez(
            path,
            prior=self.prior,
            emission=self.emission,
            transition=self.transition,
            labels=self.labels,
            ignore_transition_gaps=self.ignore_transition_gaps,
            handle_sleep_transitions=self.handle_sleep_transitions
        )

    def load(self, path):
        """
        Load model parameters from a Numpy npz file.

        :param path: npz file location
        :type path: str
        """
        d = np.load(path, allow_pickle=True)
        self.prior = d["prior"]
        self.emission = d["emission"]
        self.transition = d["transition"]
        self.labels = d["labels"]
        self.ignore_transition_gaps = d["ignore_transition_gaps"]
        self.handle_sleep_transitions = d["handle_sleep_transitions"]

    def display(self, labels, precision=3):
        """
        Print the model parameters in a readable format.

        :param labels: List of label names in expected order
        :type labels: list/dict
        :param precision: Number of decimal places to print
        :type precision: int
        """

        pretty_hmm_params(self, labels, precision)


def check_for_input_errors(Y, T, interval, groups=None, ignore_transition_gaps=False, handle_sleep_transitions=False):
    if not ignore_transition_gaps:
        if len(Y) != len(T):
            raise Exception("Provided times should have same length as labels")
        if not interval:
            raise Exception("A window length must be provided when using label times to train hmm")

    if handle_sleep_transitions:
        if len(Y) != len(groups):
            raise Exception("Provided group labels should have same length as labels")


def restore_labels_after_gaps(y_pred, y_smooth, t, interval):
    df = pd.DataFrame({"y_pred": y_pred, "y_smooth": y_smooth})

    if type(t[0]) == int:
        gaps = pd.Series(t).diff(periods=1) != interval
        gaps[0] = False
    else:
        gaps = pd.Series(t).diff(periods=1) != pd.Timedelta(seconds=interval)
        gaps[0] = False

    df.loc[gaps, "y_smooth"] = df.loc[gaps, "y_pred"]

    return df["y_smooth"].values


def calculate_transition_matrix(Y, groups=None, t=None, interval=None,
                                 ignore_transition_gaps=False, handle_sleep_transitions=False):
    check_for_input_errors(Y, t, interval, groups, ignore_transition_gaps, handle_sleep_transitions)

    if ignore_transition_gaps:
        t = range(len(Y))
        interval = 1

    df = pd.DataFrame({
        "label": Y,
        "group": groups
    })

    # create a new column with data shifted one space
    df["shift"] = df["label"].shift(-1)

    # only consider transitions of expected interval
    if type(t[0]) == int:
        df = df[(-pd.Series(t).diff(periods=-1) == interval)]
    else:
        df = df[(-pd.Series(t).diff(periods=-1) == pd.Timedelta(seconds=interval))]

    # correct sleep transitions if required
    if handle_sleep_transitions:
        df = add_sleep_sedentary_transitions(df)

    # add a count column (for group by function)
    df["count"] = 1

    # groupby and then unstack, fill the zeros
    trans_mat = df.groupby(["label", "shift"])["count"].count().unstack().fillna(0)

    # normalise by occurences and save values to get the transition matrix
    trans_mat = trans_mat.div(trans_mat.sum(axis=1), axis=0).values

    if trans_mat.size == 0:
        raise Exception("No transitions found in data")

    return trans_mat


def get_activity_label_code(label, labels):
    try:
        return list(sorted(labels)).index(label)
    except ValueError:
        raise ValueError(f"Label '{label}' not recognised. Must be one of {list(sorted(labels))}.")
    

def print_array(arr, precision=3):
    """Prints all elements of a NumPy array to N decimal places."""
    arr = np.array(arr)
    with np.printoptions(precision=precision, suppress=True):
        print(arr)


def reorder_matrix(data, index_order):
    """Reorder a 1D or 2D square array"""
    arr = np.array(data)

    if arr.ndim == 1:
        return arr[index_order]
    
    elif arr.ndim == 2:
        if arr.shape[0] != arr.shape[1]:
            raise ValueError("2D input must be a square matrix.")
        # Reorder rows and columns using the same index order
        return arr[index_order][:, index_order]
    
    else:
        raise ValueError("Input must be a 1D or 2D array.")


def pretty_hmm_params(hmm: HMM, labels, precision=3):
    """Print the HMM parameters in a readable format, reordering them according to the provided labels."""

    if isinstance(labels, dict):
        labels = list(labels.keys())

    elif not isinstance(labels, list):
        raise ValueError("labels must be a list or dict")

    index_order = [get_activity_label_code(label, labels) for label in labels]

    print(f"HMM Parameters ordered {', '.join(labels)}:\n")

    print("Prior:")
    prior = reorder_matrix(hmm.prior, index_order)
    print_array(prior, precision)

    print("\nEmission:")
    emission = reorder_matrix(hmm.emission, index_order)
    print_array(emission, precision)

    print("\nTransition:")
    transition = reorder_matrix(hmm.transition, index_order)
    print_array(transition, precision)



import os
import pathlib
import urllib
import shutil
import time
import argparse
import json
import hashlib
import warnings
import numpy as np
import pandas as pd
import joblib
import sys

import actipy

from actinet import __version__
from actinet import __classifiers__
from actinet.accPlot import plotTimeSeries
from actinet.models import ActivityClassifier
from actinet.summarisation import get_activity_summary
from actinet.utils.summary_utils import calculate_daily_wear_stats
from actinet.utils.utils import infer_freq, drop_first_last_days, flag_wear_below_days, calculate_wear_stats

BASE_URL = "https://wearables-files.ndph.ox.ac.uk/files/models/actinet/"


def main():
    parser = argparse.ArgumentParser(
        description="A tool to predict activities from accelerometer data using a self-supervised ResNet-18 model",
        add_help=True,
    )
    parser.add_argument(
        "filepath", nargs="?", default="", help="Enter file to be processed"
    )
    parser.add_argument(
        "--outdir",
        "-o",
        help="Enter folder location to save output files",
        default="outputs/",
    )
    parser.add_argument(
        "--classifier",
        "-c",
        help="Enter custom activity classifier file to use. Default: walmsley (Walmsley2020 annotations of activity intensity). Can also enter path to local classifier (.joblib.lzma) file.",
        default="walmsley",
    )
    parser.add_argument(
        "--no-hmm",
        action="store_true",
        help="Disable HMM post-processing",
    )
    parser.add_argument(
        "--require-sleep-above",
        help="Require sleep blocks to exceed a minimum duration, otherwise be classified as sedentary. Pass values as strings, e.g.: '2H', '30min'. Default: None (no requirement)",
        type=str, 
        default=None
    )
    parser.add_argument(
        "--single-sleep-block",
        action="store_true",
        help="Recognise only one sleep block per day, all other sleep blocks will be converted to sedentary",
    )
    parser.add_argument(
        "--force-download",
        action="store_true",
        help="Force download of classifier file",
    )
    parser.add_argument(
        "--pytorch-device",
        "-d",
        help="Pytorch device to use, e.g.: 'cpu' or 'cuda:0'",
        type=str,
        default="cpu",
    )
    parser.add_argument(
        "--sample-rate",
        "-r",
        help="Sample rate for measurement, otherwise inferred.",
        type=int,
        default=None,
    )
    parser.add_argument(
        "--exclude-first-last", 
        "-e",
        help="Exclude first, last or both days of data. Default: None (no exclusion)",
        type=str, 
        choices=['first', 'last', 'both'], 
        default=None
    ) 
    parser.add_argument(
        "--exclude-wear-below", 
        "-w",
        help="Exclude days with wear time below threshold. Pass values as strings, e.g.: '12H', '30min'. Default: None (no exclusion)",
        type=str, 
        default=None)
    parser.add_argument(
        "--csv-start-row",
        help="Row number to start reading a CSV file. Default: 1 (First row)",
        type=int,
        default=1,
    )
    parser.add_argument("--csv-txyz",
                        help="Column names for time, x, y, z in CSV files. "
                             "Comma-separated string. Default: 'time,x,y,z'",
                        type=str, default="time,x,y,z"
    )
    parser.add_argument("--csv-txyz-idxs",
                        help="Column indices for time,x,y,z (0-indexed, e.g., '0,1,2,3'). Overrides --csv-txyz.",
                        type=str, default=None)
    parser.add_argument('--csv-date-format',
                        default="%Y-%m-%d %H:%M:%S.%f",
                        type=str, 
                        help="Date time format for csv file when reading a csv file. " +
                             "See https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes for more possible codes." +
                             "Default: '%%Y-%%m-%%d %%H:%%M:%%S.%%f' (e.g. '2023-10-01 12:34:56.789')"
    )
    parser.add_argument('--calibration-stdtol-min',
                        default=None,
                        type=float,
                        help="Minimum standard deviation tolerance (g) for detecting stationary periods for calibration. Default: None"
    )
    parser.add_argument(
        "--plot-activity",
        "-p",
        action="store_true",
        help="Plot the predicted activity labels",
    )
    parser.add_argument(
        "--cache-classifier",
        action="store_true",
        help="Download and cache classifier file and model modules for offline usage",
    )
    parser.add_argument(
        "--model-repo-path", "-m", help="Enter repository of ssl model", default=None
    )
    parser.add_argument("--quiet", "-q", action="store_true", help="Suppress output")
    args = parser.parse_args()

    before = time.time()

    verbose = not args.quiet

    if args.cache_classifier:
        load_classifier(
            classifier=args.classifier,
            model_repo_path=None,
            force_download=True,
            verbose=verbose,
        )

        after = time.time()
        print(f"Done! ({round(after - before, 2)}s)")

        return

    else:
        if not args.filepath:
            raise ValueError("Please provide a file to process.")

    # Info contains high-level summary of the data and results
    info = {}
    info['ActiNetVersion'] = __version__
    info['ActiNetArgs'] = vars(args)

    # Load file
    data, info_read = read(
        args.filepath,
        args.csv_txyz,
        args.csv_start_row-1,  # -1 to convert to zero-based index
        args.csv_date_format,
        args.calibration_stdtol_min,
        resample_hz="uniform",
        sample_rate=args.sample_rate,
        csv_txyz_idxs=args.csv_txyz_idxs,
        verbose=verbose,
    )
    info.update(info_read)

    # Exclusion: first/last days
    if args.exclude_first_last is not None:
        data = drop_first_last_days(data, args.exclude_first_last)
  
    # Exclusion: days with wear time below threshold
    if args.exclude_wear_below is not None:
        data = flag_wear_below_days(data, args.exclude_wear_below)
   
    # Update wear time stats after exclusions
    info.update(calculate_wear_stats(data))

    daily_wear_stats = calculate_daily_wear_stats(data)

    # Output paths
    basename = resolve_path(args.filepath)[1]
    outdir = os.path.join(args.outdir, basename)
    outputSummaryFile = f"{outdir}/{basename}-outputSummary.json"
    os.makedirs(outdir, exist_ok=True)

     # If no data, save info and exit
    if len(data) == 0 or data.isna().any(axis=1).all():  # TODO: check na only on x,y,z cols?
        # Save info as outputSummary.json
        with open(outputSummaryFile, "w") as f:
            json.dump(info, f, indent=4, cls=NpEncoder)

        # Print
        if verbose:
            print("\nSummary Stats\n---------------------")
            print(
                json.dumps(
                    {
                        key: info[key]
                        for key in [
                            "Filename",
                            "Filesize(MB)",
                            "WearTime(days)",
                            "NonwearTime(days)",
                            "ReadOK"
                        ]
                    },
                    indent=4,
                    cls=NpEncoder,
                )
            )
        print("No data to process. Exiting early...")
        sys.exit(0)

    # Run classifier
    if verbose:
        print("Loading classifier...")

    classifier: ActivityClassifier = load_classifier(
        args.classifier,
        args.model_repo_path,
        args.force_download,
        verbose
    )

    classifier.verbose = verbose
    classifier.device = args.pytorch_device

    if verbose:
        print("Running activity classifier...")
    Y = classifier.predict_from_frame(data, args.sample_rate, not args.no_hmm,
                                      args.require_sleep_above, args.single_sleep_block)

    # Save predicted activities
    timeSeriesFile = f"{outdir}/{basename}-timeSeries.csv.gz"
    Y.to_csv(timeSeriesFile)

    if verbose:
        print("Time series output written to:", timeSeriesFile)

    # Plot activity profile
    if args.plot_activity:
        plotFile = f"{outdir}/{basename}-timeSeries-plot.png"
        fig = plotTimeSeries(Y)
        fig.savefig(plotFile, dpi=200, bbox_inches="tight")

        if verbose:
            print("Output plot written to:", plotFile)

    # Summary
    summary, daily_summary = get_activity_summary(Y, list(classifier.labels), args.exclude_wear_below,
                                                  True, True, verbose)

    # Join the actipy processing info, with acitivity summary data
    outputSummary = {**summary, **info}

    # Save output summary
    with open(outputSummaryFile, "w") as f:
        json.dump(outputSummary, f, indent=4, cls=NpEncoder)

    if verbose:
        print("Output summary written to:", outputSummaryFile)

    daily_summary = pd.concat([daily_wear_stats, daily_summary], axis=1)

    daily_summary.insert(0, 'Filename', info['Filename'])  # add filename for reference
    daily_summary.to_csv(f"{outdir}/{basename}-Daily.csv.gz")

    # Print
    if verbose:
        print("\nSummary Stats\n---------------------")
        print(
            json.dumps(
                {
                    key: outputSummary[key]
                    for key in [
                        "Filename",
                        "Filesize(MB)",
                        "WearTime(days)",
                        "NonwearTime(days)",
                        "ReadOK",
                    ]
                    + [
                        f"{label}-overall-avg"
                        for label in ["acc"] + list(classifier.labels)
                    ]
                },
                indent=4,
                cls=NpEncoder,
            )
        )

    after = time.time()
    print(f"Done! ({round(after - before,2)}s)")


def read(
    filepath, usecols=None, skipRows=0, dateFormat=None, calibration_stdtol_min=None,
    resample_hz="uniform", sample_rate=None, lowpass_hz=None, csv_txyz_idxs=None, verbose=True
):
    p = pathlib.Path(filepath)
    fsize = round(p.stat().st_size / (1024 * 1024), 1)
    
    ftype = p.suffix.lower()
    if ftype in (".gz", ".xz", ".lzma", ".bz2", ".zip"):  # if file is compressed, check the next extension
        ftype = pathlib.Path(p.stem).suffix.lower()

    if ftype in (".csv", ".pkl"):
        if ftype == ".csv":
            # Determine column names: either from indices or from usecols
            if csv_txyz_idxs is not None:
                try:
                    tidx, xidx, yidx, zidx = map(int, csv_txyz_idxs.split(','))
                except ValueError:
                    raise ValueError(f"csv_txyz_idxs must be 4 comma-separated integers, got: '{csv_txyz_idxs}'")
                if any(i < 0 for i in [tidx, xidx, yidx, zidx]):
                    raise ValueError(f"csv_txyz_idxs must be non-negative integers, got: '{csv_txyz_idxs}'")
                header = pd.read_csv(filepath, nrows=0).columns.tolist()
                max_idx = max(tidx, xidx, yidx, zidx)
                if max_idx >= len(header):
                    raise ValueError(f"Column index {max_idx} out of range. CSV has {len(header)} columns.")
                tcol, xcol, ycol, zcol = header[tidx], header[xidx], header[yidx], header[zidx]
            else:
                tcol, xcol, ycol, zcol = usecols.split(',')
            
            data = pd.read_csv(
                filepath,
                usecols=[tcol, xcol, ycol, zcol],
                parse_dates=[tcol],
                date_format=dateFormat,
                index_col=tcol,
                dtype={xcol: "f4", ycol: "f4", zcol: "f4"},
                skiprows=skipRows,
            )

            # rename to standard names
            data = data.rename(columns={xcol: 'x', ycol: 'y', zcol: 'z'})
            data.index.name = 'time'

        elif ftype == ".pkl":
            data = pd.read_pickle(filepath)

        if sample_rate in (None, False):
            freq = infer_freq(data.index)
            sample_rate = int(np.round(pd.Timedelta("1s") / freq))

        # Quick fix: Drop duplicate indices. TODO: Maybe should be handled by actipy.
        data = data[~data.index.duplicated(keep="first")]

        data, info = actipy.process(
            data,
            sample_rate,
            lowpass_hz=None,
            calibrate_gravity=True,
            calibrate_gravity_kwargs={'stdtol_min': calibration_stdtol_min},
            detect_nonwear=True,
            resample_hz=resample_hz,
            verbose=verbose,
        )

        info = {
            **{
                "Filename": filepath,
                "Device": ftype,
                "Filesize(MB)": fsize,
                "SampleRate": sample_rate,
                "ReadOK": 1,
            },
            **info,
        }

    elif ftype in (".cwa", ".gt3x", ".bin"):

        if csv_txyz_idxs is not None:
            warnings.warn("--csv-* options are only supported for CSV files. Ignoring.")

        data, info = actipy.read_device(
            filepath,
            lowpass_hz=None,
            calibrate_gravity=True,
            calibrate_gravity_kwargs={'stdtol_min': calibration_stdtol_min},
            detect_nonwear=True,
            resample_hz=resample_hz,
            verbose=verbose,
        )

    else:
        raise ValueError(f"Unknown file format: {ftype}")

    if "ResampleRate" not in info:
        info["ResampleRate"] = info["SampleRate"]

    return data, info


def resolve_path(path):
    """Return parent folder, file name and file extension"""
    p = pathlib.Path(path)
    extension = p.suffixes[0]
    filename = p.name.rsplit(extension)[0]
    dirname = p.parent
    return dirname, filename, extension


def load_classifier(
    classifier,
    model_repo_path=None,
    force_download=False,
    verbose=True,
):
    """Load trained classifier. Download if not exists."""

    if classifier in __classifiers__.keys():
        classifier_version = __classifiers__[classifier]['version']
        classifier_md5 = __classifiers__[classifier]['md5']
        
        classifier_path = pathlib.Path(__file__).parent / f"{classifier_version}.joblib.lzma"

        if force_download or not classifier_path.exists():
            url = f"{BASE_URL}{classifier_version}.joblib.lzma"

            if verbose:
                print(f"Downloading {url}...")

            with urllib.request.urlopen(url) as f_src, open(classifier_path, "wb") as f_dst:
                shutil.copyfileobj(f_src, f_dst)

            if classifier_md5:
                digest = md5(classifier_path)
                if digest != classifier_md5:
                    raise ValueError(
                        "Classifier file is corrupted. "
                        "Run again with --force-download to redownload."
                    )

    elif pathlib.Path(classifier).exists():
        classifier_path = classifier

    else:
        raise ValueError(f"Unknown classifier: {classifier}")

    try:
        classifier: ActivityClassifier = joblib.load(classifier_path)
    except Exception as e:
        raise ValueError(f"Error loading classifier file.")

    if model_repo_path and pathlib.Path(model_repo_path).exists() and verbose:
        print(f"Loading model repository from {model_repo_path}.")

    classifier.load_model(model_repo_path)

    return classifier


def md5(fname):
    hash_md5 = hashlib.md5()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


class NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if pd.isnull(obj):  # handles pandas NAType
            return np.nan
        return json.JSONEncoder.default(self, obj)


if __name__ == "__main__":
    main()



import numpy as np
import pandas as pd
import copy
import os
import joblib
import torch
from tqdm.auto import tqdm
from torch.utils.data import DataLoader
from sklearn.model_selection import StratifiedGroupKFold, GroupShuffleSplit
from imblearn.ensemble import BalancedRandomForestClassifier
from sklearn.preprocessing import LabelEncoder
from scipy.special import softmax

from actinet import hmm
from actinet import sslmodel
from actinet.utils.utils import safe_indexer, resize, infer_freq
from actinet.utils.sleep_utils import removeSpuriousSleep


class ActivityClassifier:
    """
    Implement a ResNet-18 based Activity Classifier with model saving/loading and optional HMM smoothing.
    """
    def __init__(
        self,
        device="cpu",
        batch_size=512,
        window_sec=30,
        weights_path=None,
        labels=[],
        repo_tag="v1.0.0",
        hmm_params=None,
        hmm_ignore_transition_gaps=False,
        hmm_handle_sleep_transitions=False,
        verbose=False,
    ):
        self.device = device
        self.repo_tag = repo_tag
        self.batch_size = batch_size
        self.window_sec = window_sec
        self.labels = labels
        self.verbose = verbose

        self.model_weights = (
            sslmodel.get_model_dict(weights_path, device) if weights_path else None
        )
        self.model = None

        self.hmm = load_hmm_params(hmm_params, 
                                   hmm_ignore_transition_gaps,
                                   hmm_handle_sleep_transitions,
                                   verbose)

    def __str__(self):
        return (
            "Activity Classifier\n"
            "class_labels: {self.labels}\n"
            "window_length: {self.window_sec}\n"
            "batch_size: {self.batch_size}\n"
            "device: {self.device}\n"
            "hmm: {self.hmm}\n"
            "model: {model}".format(
                self=self, model=self.model or "Model has not been loaded."
            )
        )

    def fit(
        self,
        X,
        Y,
        groups=None,
        T=None,
        weights_path="models/weights.pt",
        model_repo_path=None,
        n_splits=5,
    ):
        """
        Fit the ActivityClassifier to the provided data by training the model.

        :param X: The training accelerometer data [x,y,z] with shape (rows, window_len, 3)
        :type X: numpy.ndarray
        :param Y: Ground truth labels of the training data with shape (rows, )
        :type Y: numpy.ndarray
        :param groups: Participant labels for the training data with shape (rows, )
        :type groups: numpy.ndarray, optional
        :param T: Time at each observation with shape (rows, )
        :type T: numpy.ndarray, optional
        :param weights_path: Path to save the model weights
        :type weights_path: str, optional
        :param model_repo_path: Path to the ssl-wearables model repository (https://github.com/OxWearables/ssl-wearables)
        :type model_repo_path: str, optional
        :param n_splits: Number of splits for cross-validation
        :type n_splits: int, optional
        """
        sslmodel.verbose = self.verbose

        Y = LabelEncoder().fit_transform(Y)

        if self.verbose:
            print("Training SSL")

        y_prob_splits = []
        y_true_splits = []
        group_splits = []
        t_splits = []

        if n_splits < 3:
            splitter = GroupShuffleSplit(n_splits=n_splits, random_state=42)
            split_iterator = splitter.split(X, Y, groups)
        else:
            splitter = StratifiedGroupKFold(n_splits)
            split_iterator = splitter.split(X, Y, groups)

        for i, (train_idx, val_idx) in enumerate(split_iterator):
            if self.verbose:
                print(f"Training split {i+1}/{n_splits}")

            x_train = X[train_idx]
            x_val = X[val_idx]

            y_train = Y[train_idx]
            y_val = Y[val_idx]

            group_train = safe_indexer(groups, train_idx)
            group_val = safe_indexer(groups, val_idx)

            t_val = safe_indexer(T, val_idx)

            train_dataset = sslmodel.NormalDataset(
                x_train, y_train, pid=group_train, augmentation=True
            )
            val_dataset = sslmodel.NormalDataset(x_val, y_val, pid=group_val)

            train_loader = DataLoader(
                train_dataset,
                batch_size=self.batch_size,
                shuffle=True,
                num_workers=1,
            )

            val_loader = DataLoader(
                val_dataset,
                batch_size=self.batch_size,
                shuffle=False,
                num_workers=1,
            )

            self.load_model(model_repo_path)

            if self.model_weights is None or not os.path.exists(weights_path):
                sslmodel.train(
                    self.model,
                    train_loader,
                    val_loader,
                    self.device,
                    weights_path=weights_path,
                    class_weights="balanced",
                )
                self.model.load_state_dict(torch.load(weights_path, self.device))

            # train HMM with predictions of the validation set
            y_val, y_val_pred, _ = sslmodel.predict(
                self.model, val_loader, self.device, output_logits=True
            )
            y_val_pred_sf = softmax(y_val_pred, axis=1)

            y_true_splits.append(y_val)
            y_prob_splits.append(y_val_pred_sf)
            group_splits.append(group_val)
            t_splits.append(t_val)

        y_prob_splits = np.vstack(y_prob_splits)
        y_true_splits = np.hstack(y_true_splits)
        group_splits = np.hstack(group_splits)
        t_splits = np.hstack(t_splits)

        if self.verbose:
            print("Training HMM")

        self.hmm.fit(
            y_prob_splits,
            y_true_splits,
            group_splits,
            t_splits,
            interval=self.window_sec
        )

        # move model to cpu to get a device-less state dict (prevents device conflicts when loading on cpu/gpu later)
        self.model.to("cpu")
        self.model_weights = self.model.state_dict()

        return self

    def predict_from_frame(self, data, sample_freq, hmm_smothing=True, sleep_tolerance=None, remove_naps=False):
        """
        Use the ActivityClassifier to make predictions on input accelerometer data.

        :param data: The training accelerometer data [x,y,z]
        :type data: pandas.DataFrame
        :param sample_freq: Sampling frequency of the accelerometer data
        :type sample_freq: int or float
        :param hmm_smothing: Whether to apply HMM smoothing to the predictions
        :type hmm_smothing: bool, optional
        :param sleep_tolerance: Time threshold for sleep periods to be considered valid (e.g., '30min')
        :type sleep_tolerance: str, optional
        :param remove_naps: Whether to remove nap periods from the predictions
        :type remove_naps: bool, optional
        """
        sample_freq = sample_freq or 1 / (infer_freq(data.index).total_seconds())
        X, T = make_windows(
            data,
            self.window_sec,
            int(self.window_sec * sample_freq),
            return_index=True,
            verbose=self.verbose,
        )

        Y = raw_to_df(
            X, self.predict(X, T, hmm_smothing, sleep_tolerance, remove_naps), 
            T, self.labels, reindex=False
        )

        return Y

    def predict(self, X, T=None, hmm_smothing=True, sleep_tol=None, remove_naps=False):
        """
        Use the ActivityClassifier to make predictions on input accelerometer data.

        :param X: The input accelerometer data [x,y,z] with shape (rows, window_len, 3)
        :type X: numpy.ndarray
        :param T: Time at each observation with shape (rows, )
        :type T: numpy.ndarray, optional
        :param hmm_smothing: Whether to apply HMM smoothing to the predictions
        :type hmm_smothing: bool, optional
        :param sleep_tol: Time threshold for sleep periods to be considered valid (e.g., '30min')
        :type sleep_tol: str, optional
        """
        if self.model is None:
            raise Exception("Model has not been loaded for ActivityClassifier.")

        self.model.to(self.device)

        # check X quality
        ok = np.flatnonzero(~np.asarray([np.isnan(x).any() for x in X]))

        X_ = X[ok]
        T_ = safe_indexer(T, ok)

        sslmodel.verbose = self.verbose

        dataset = sslmodel.NormalDataset(X_)
        dataloader = DataLoader(
            dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=0,
        )

        _, Y_, _ = sslmodel.predict(
            self.model, dataloader, self.device, output_logits=False
        )

        if hmm_smothing:
            interval = self.window_sec if T_ is not None else None
            Y_ = self.hmm.predict(Y_, T_, interval)

        Y = np.full(len(X), fill_value=np.nan)
        Y[ok] = Y_

        Y = removeSpuriousSleep(Y, self.labels, self.window_sec, sleep_tol, remove_naps)

        return Y

    def load_model(self, model_repo_path=None):
        """
        Load SSL model reposiotory from specified path. (https://github.com/OxWearables/ssl-wearables)

        :param model_repo_path: Path to the ssl-wearables model repository
        :type model_repo_path: str, optional
        """
        self.model = sslmodel.get_sslnet(
            tag=self.repo_tag,
            local_repo_path=model_repo_path,
            pretrained_weights=self.model_weights or True,
            window_sec=self.window_sec,
            num_labels=len(self.labels),
        )
        self.model.to(self.device)

    def save(self, output_path):
        """
        Save the ActivityClassifier model to a .lzma file.

        :param output_path: lzma file location
        :type output_path: str
        """
        classifier = copy.deepcopy(self)
        classifier.model = None
        classifier.device = "cpu"
        classifier.batch_size = 512

        joblib.dump(classifier, output_path, compress=("lzma", 3))


class RFActivityClassifier:
    """
    Implement a basic Balanced Random Forest classifier with optional HMM smoothing.
    """

    def __init__(
        self, 
        winsec=None, 
        hmm_params=None, 
        hmm_ignore_transition_gaps=False,
        hmm_handle_sleep_transitions=False,
        labels=None,
        verbose=False,
        **kwargs):

        self.model = BalancedRandomForestClassifier(oob_score=True, random_state=42, verbose=verbose, **kwargs)
        self.labels = labels
        self.hmm = load_hmm_params(hmm_params, hmm_ignore_transition_gaps, hmm_handle_sleep_transitions, verbose)
        self.winsec = winsec

    def __str__(self):
        return str(self.model)

    def fit(self, X, Y, groups=None, T=None):
        self.model.fit(X, Y)
        self.hmm.fit(self.model.oob_decision_function_, Y, groups, T, self.winsec)

    def predict(self, X, T=None, hmm_smothing=True, sleep_tol=None, remove_naps=False):
        y_pred = self.model.predict(X)

        if hmm_smothing:
            y_pred = self.hmm.predict(y_pred, T, self.winsec)

        y_pred = removeSpuriousSleep(y_pred, self.labels, self.winsec, sleep_tol, remove_naps)

        return y_pred
    
    def save(self, output_path):
        classifier = copy.deepcopy(self)

        joblib.dump(classifier, output_path, compress=("lzma", 3))

    def load(self, input_path):
        classifier = joblib.load(input_path)
        self.model = classifier.model
        self.labels = classifier.labels
        self.hmm = classifier.hmm
        self.winsec = classifier.winsec


def make_windows(data, window_sec, window_len, return_index=False, verbose=True):
    """Split data into windows"""

    if verbose:
        print("Defining windows...")

    X, T = [], []
    acc_cols = ["x", "y", "z"]
    ssl_window_len = int(sslmodel.SAMPLE_RATE * window_sec)

    for t, x in tqdm(
        data.resample(f"{window_sec}s", origin="start"),
        mininterval=5,
        disable=not verbose,
    ):
        n = len(x)
        x = x[acc_cols].to_numpy()

        if n == window_len:
            x = x
        elif n > window_len:
            x = x[:window_len]
        elif n < window_len and n > window_len / 2:
            x = np.pad(x, ((0, window_len - n), (0, 0)), mode="wrap")
        else:
            x = np.full((window_len, 3), np.nan)

        X.append(x)
        T.append(t)

    X = np.asarray(X)

    if window_len != ssl_window_len:
        X = resize(X, ssl_window_len)

    if return_index:
        T = pd.DatetimeIndex(T, name=data.index.name)
        return X, T

    return X


def raw_to_df(data, labels, time, classes, reindex=True, freq="30S"):
    """
    Construct a DataFrome from the raw data, prediction labels and time Numpy arrays.

    :param data: Numpy windowed acc data, shape (rows, window_len, 3)
    :param labels: Either a scalar label array with shape (rows, ),
                    or the probabilities for each class if label_proba==True with shape (rows, len(classes)).
    :param time: Numpy time array, shape (rows, )
    :param classes: Array with the categorical class labels.
                    The index of this array should correspond to the labels value when label_proba==False.
    :param reindex: Reindex the dataframe to fill missing values
    :param freq: Reindex frequency
    :return: Dataframe
        Index: DatetimeIndex
        Columns: acc, classes
    :rtype: pd.DataFrame
    """
    label_matrix = np.zeros((len(time), len(classes)), dtype=np.float32)
    a_matrix = np.zeros(len(time), dtype=np.float32)

    for i, data in enumerate(data):
        if np.isnan(labels[i]):
            label_matrix[i, :] = np.nan
            a_matrix[i] = np.nan
            continue

        label = int(labels[i])
        label_matrix[i, label] = 1

        x = data[:, 0]
        y = data[:, 1]
        z = data[:, 2]

        enmo = (np.sqrt(x**2 + y**2 + z**2) - 1) * 1000  # in milli gravity
        enmo[enmo < 0] = 0
        a_matrix[i] = np.mean(enmo)

    datadict = {
        **{"time": time, "acc": a_matrix},
        **{classes[i]: label_matrix[:, i] for i in range(len(classes))},
    }

    df = pd.DataFrame(datadict)
    df = df.set_index("time")

    if reindex:
        newindex = pd.date_range(df.index[0], df.index[-1], freq=freq)
        df = df.reindex(newindex, method="nearest", fill_value=np.nan, tolerance="5S")

    return df


def load_hmm_params(hmm_params, ignore_transition_gaps, handle_sleep_transitions, verbose=False):
    if isinstance(hmm_params, str):
        if os.path.exists(hmm_params):
            if verbose:
                print(f"Loading hmm_params from {hmm_params}")

            hmm_params = dict(np.load(hmm_params, allow_pickle=True))

        else:
            raise FileNotFoundError(
                "Path to file with saved hmm parameters cannot be found."
            )

    elif hmm_params is None:
        hmm_params = dict()

    elif not isinstance(hmm_params, dict):
        raise TypeError(
            "Invalid type for HMM parameters. Expected str, dict, or None."
        )
    
    hmm_params.update({
        "ignore_transition_gaps": ignore_transition_gaps,
        "handle_sleep_transitions": handle_sleep_transitions
    })

    return hmm.HMM(**hmm_params)



""" Helper classes and functions for the SSL model """

from collections import OrderedDict
import torch
import torch.nn as nn
import numpy as np
import random
from pathlib import Path
from transforms3d.axangles import axangle2mat
from tqdm import tqdm
import warnings
with warnings.catch_warnings():
    warnings.filterwarnings("ignore", message="Failed to load image Python extension")
    from torchvision import transforms
from torch.utils.data.dataset import Dataset

SAMPLE_RATE = 30  # Sample rate for SSL Resnet18 model fixed to 30Hz
verbose = True
torch_cache_path = Path(__file__).parent / "torch_hub_cache"

torch.multiprocessing.set_sharing_strategy("file_system")


class RandomSwitchAxis:
    """
    Randomly switch the three axises for the raw files
    Input size: 3 * FEATURE_SIZE
    """

    def __call__(self, sample):
        # 3 * FEATURE
        x = sample[0, :]
        y = sample[1, :]
        z = sample[2, :]

        choice = random.randint(1, 6)

        if choice == 1:
            sample = torch.stack([x, y, z], dim=0)
        elif choice == 2:
            sample = torch.stack([x, z, y], dim=0)
        elif choice == 3:
            sample = torch.stack([y, x, z], dim=0)
        elif choice == 4:
            sample = torch.stack([y, z, x], dim=0)
        elif choice == 5:
            sample = torch.stack([z, x, y], dim=0)
        elif choice == 6:
            sample = torch.stack([z, y, x], dim=0)

        return sample


class RotationAxis:
    """
    Rotation along an axis
    """

    def __call__(self, sample):
        # 3 * FEATURE_SIZE
        sample = np.swapaxes(sample, 0, 1)
        angle = np.random.uniform(low=-np.pi, high=np.pi)
        axis = np.random.uniform(low=-1, high=1, size=sample.shape[1])
        sample = np.matmul(sample, axangle2mat(axis, angle))
        sample = np.swapaxes(sample, 0, 1)
        return sample


class RandomDecimation:
    """
    Randomly decimate the input along the time axis
    """

    def __call__(self, sample):
        decimation_factor = random.randint(1, 3)
        T = sample.shape[1]
        sample = sample[:, ::decimation_factor]
        
        sample = torch.nn.functional.interpolate(
            sample.unsqueeze(0),
            size=T,
            mode="linear",
            align_corners=False,
        ).squeeze(0)
        return sample


class NormalDataset(Dataset):
    def __init__(
        self, X, y=None, pid=None, augmentation=False, transpose_channels_first=True
    ):

        X = X.astype("f4")  # PyTorch defaults to float32

        if transpose_channels_first:
            X = np.transpose(X, (0, 2, 1))
        self.X = torch.from_numpy(X)

        if y is not None:
            self.y = torch.tensor(y)
        else:
            self.y = None

        self.pid = pid

        if augmentation:
            self.transform = transforms.Compose([RandomSwitchAxis(), RotationAxis(), RandomDecimation()])
        else:
            self.transform = None

    def __len__(self):
        return len(self.X)

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()

        sample = self.X[idx, :]

        if self.y is not None:
            y = self.y[idx]
        else:
            y = np.NaN

        if self.pid is not None:
            pid = self.pid[idx]
        else:
            pid = np.NaN

        if self.transform is not None:
            sample = self.transform(sample)

        return sample, y, pid


class EarlyStopping:
    """Early stops the training if validation loss
    doesn't improve after a given patience."""

    def __init__(
        self,
        patience=5,
        verbose=False,
        delta=0,
        path="checkpoint.pt",
        trace_func=print,
    ):
        """
        Args:
            patience (int): How long to wait after last time v
                            alidation loss improved.
                            Default: 7
            verbose (bool): If True, prints a message for each
                            validation loss improvement.
                            Default: False
            delta (float): Minimum change in the monitored quantity
                            to qualify as an improvement.
                            Default: 0
            path (str): Path for the checkpoint to be saved to.
                            Default: 'checkpoint.pt'
            trace_func (function): trace print function.
                            Default: print
        """
        self.patience = patience
        self.verbose = verbose
        self.counter = 0
        self.best_score = None
        self.early_stop = False
        self.val_loss_min = np.Inf
        self.delta = delta
        self.trace_func = trace_func

        self.path = path

    def __call__(self, val_loss, model):

        score = -val_loss

        if self.best_score is None:
            self.best_score = score
            self.save_checkpoint(val_loss, model)
        elif score < self.best_score + self.delta:
            self.counter += 1
            if self.verbose:
                self.trace_func(
                    f"EarlyStopping counter: {self.counter}/{self.patience}"
                )
            if self.counter >= self.patience:
                self.early_stop = True
        else:
            self.best_score = score
            self.save_checkpoint(val_loss, model)
            self.counter = 0

    def save_checkpoint(self, val_loss, model):
        """Saves model when validation loss decrease."""
        if self.verbose:
            msg = "Validation loss decreased"
            msg = msg + f" ({self.val_loss_min:.6f} --> {val_loss:.6f}). "
            msg = msg + "Saving model ..."
            self.trace_func(msg)
        if hasattr(model, "module"):
            torch.save(model.module.state_dict(), self.path)
        else:
            torch.save(model.state_dict(), self.path)
        self.val_loss_min = val_loss


def get_sslnet(
    tag="v1.0.0",
    local_repo_path=None,
    pretrained_weights=False,
    window_sec: int = 30,
    num_labels: int = 4,
):
    """
    Load and return the Self Supervised Learning (SSL) model from pytorch hub or local storage.

    :param str device: PyTorch device to use
    :param str tag: Tag on the ssl-wearables repo to check out
    :param str local_repo_path: Path to local version of the SSL repository for offline usage
    :param bool/OrderedDict pretrained_weights: Initialise the model with UKB self-supervised/specified pretrained weights
    :param int window_sec: The length of the window of data in seconds (limited to 5, 10 or 30)
    :param int num_labels: The number of labels to predict
    :return: pytorch SSL model
    :rtype: nn.Module
    """

    if window_sec not in [5, 10, 30]:
        raise ValueError(
            "Length of window in seconds must be either 5, 10 or 30 seconds"
        )

    if num_labels < 1:
        raise ValueError("Numer of class labels should be > 0")

    if local_repo_path is not None:
        sslnet: nn.Module = torch.hub.load(
            local_repo_path,
            f"harnet{window_sec}",
            source="local",
            class_num=num_labels,
            pretrained=pretrained_weights == True,
        )

    else:
        repo_name = "ssl-wearables"
        repo = f"OxWearables/{repo_name}:{tag}"

        if not torch_cache_path.exists():
            Path.mkdir(torch_cache_path, parents=True, exist_ok=True)

        torch.hub.set_dir(str(torch_cache_path))

        # find repo cache dir that matches repo name and tag
        cache_dirs = [f for f in torch_cache_path.iterdir() if f.is_dir()]
        repo_path = next(
            (f for f in cache_dirs if repo_name in f.name and tag in f.name), None
        )

        if repo_path is None:
            repo_path = repo
            source = "github"
        else:
            repo_path = str(repo_path)
            source = "local"
            if verbose:
                print(f"Using local {repo_path}")

        sslnet: nn.Module = torch.hub.load(
            repo_path,
            f"harnet{window_sec}",
            trust_repo=True,
            source=source,
            class_num=num_labels,
            pretrained=pretrained_weights == True,
            verbose=verbose,
        )

    if isinstance(pretrained_weights, OrderedDict):
        sslnet.load_state_dict(pretrained_weights)

    return sslnet


def get_model_dict(weights_path, device):
    return torch.load(weights_path, map_location=device)


def predict(model, dataloader, device, output_logits=False):
    """
    Iterate over the dataloader and do prediction with a pytorch model.

    :param nn.Module model: pytorch Module
    :param DataLoader data_loader: pytorch dataloader
    :param str device: pytorch map device
    :param bool output_logits: When True, output the raw outputs (logits) from the last layer (before classification).
                                When False, argmax the logits and output a classification scalar.
    :return: true labels, model predictions, pids
    :rtype: (np.ndarray, np.ndarray, np.ndarray)
    """

    if verbose:
        print("Classifying windows...")

    predictions_list = []
    true_list = []
    pid_list = []
    model.eval()

    if len(dataloader) == 0:
        return np.array([]), np.array([]), np.array([])

    with torch.inference_mode():
        for x, y, pid in tqdm(dataloader, mininterval=5, disable=not verbose):
            x = x.to(device, dtype=torch.float)
            logits = model(x)
            true_list.append(y)
            if output_logits:
                predictions_list.append(logits.cpu())
            else:
                pred_y = torch.argmax(logits, dim=1)
                predictions_list.append(pred_y.cpu())
            pid_list.extend(pid)

    true_list = torch.cat(true_list)
    predictions_list = torch.cat(predictions_list)

    if output_logits:
        return (
            torch.flatten(true_list).numpy(),
            predictions_list.numpy(),
            np.array(pid_list),
        )
    else:
        return (
            torch.flatten(true_list).numpy(),
            torch.flatten(predictions_list).numpy(),
            np.array(pid_list),
        )


def train(
    model,
    train_loader,
    val_loader,
    device,
    class_weights=None,
    weights_path="weights.pt",
    num_epoch=100,
    learning_rate=0.0001,
    patience=5,
):
    """
    Iterate over the training dataloader and train a pytorch model.
    After each epoch, validate model and early stop when validation loss function bottoms out.

    Trained model weights will be saved to disk (weights_path).

    :param nn.Module model: pytorch model
    :param DataLoader train_loader: training data loader
    :param DataLoader val_loader: validation data loader
    :param str device: pytorch map device
    :param class_weights: Array of training class weights to use with weighted cross entropy loss.
                        Leave empty to use unweighted loss. Set to "balanced" to use inverse class weights.
    :param weights_path: save location for the trained weights (state_dict)
    :param num_epoch: number of training epochs
    :param learning_rate: Adam learning rate
    :param patience: early stopping patience
    """
    optimizer = torch.optim.Adam(model.parameters(), lr=learning_rate, amsgrad=True)

    if class_weights is not None:
        if class_weights == "balanced":
            class_weights = get_inverse_class_weights(train_loader.dataset.y.numpy())

        class_weights = torch.FloatTensor(class_weights).to(device)

        loss_fn = nn.CrossEntropyLoss(weight=class_weights)
    else:
        loss_fn = nn.CrossEntropyLoss()

    early_stopping = EarlyStopping(
        patience=patience, path=weights_path, verbose=verbose, trace_func=print
    )

    for epoch in range(num_epoch):
        model.train()
        train_losses = []
        train_acces = []
        for x, y, _ in tqdm(train_loader, disable=not verbose):
            x.requires_grad_(True)
            x = x.to(device, dtype=torch.float)
            true_y = y.to(device, dtype=torch.long)

            optimizer.zero_grad()

            logits = model(x)
            loss = loss_fn(logits, true_y)
            loss.backward()
            optimizer.step()

            pred_y = torch.argmax(logits, dim=1)
            train_acc = torch.sum(pred_y == true_y)
            train_acc = train_acc / (pred_y.size()[0])

            train_losses.append(loss.cpu().detach())
            train_acces.append(train_acc.cpu().detach())

        val_loss, val_acc = _validate_model(model, val_loader, device, loss_fn)

        epoch_len = len(str(num_epoch))
        print_msg = (
            f"[{epoch:>{epoch_len}}/{num_epoch:>{epoch_len}}] | "
            + f"train_loss: {np.mean(train_losses):.3f} | "
            + f"train_acc: {np.mean(train_acces):.3f} | "
            + f"val_loss: {val_loss:.3f} | "
            + f"val_acc: {val_acc:.2f}"
        )

        early_stopping(val_loss, model)

        if verbose:
            print(print_msg)

        if early_stopping.early_stop:
            if verbose:
                print("Early stopping")
                print(f"SSLNet weights saved to {weights_path}")
            break

    return model


def _validate_model(model, val_loader, device, loss_fn):
    """Iterate over a validation data loader and return mean model loss and accuracy."""
    model.eval()
    losses = []
    acces = []
    with torch.inference_mode():
        for x, y, _ in val_loader:
            x = x.to(device, dtype=torch.float)
            true_y = y.to(device, dtype=torch.long)

            logits = model(x)
            loss = loss_fn(logits, true_y)

            pred_y = torch.argmax(logits, dim=1)

            val_acc = torch.sum(pred_y == true_y)
            val_acc = val_acc / (list(pred_y.size())[0])

            losses.append(loss.cpu().detach())
            acces.append(val_acc.cpu().detach())
    losses = np.array(losses)
    acces = np.array(acces)
    return np.mean(losses), np.mean(acces)


def get_inverse_class_weights(y):
    """Return a list with inverse class frequencies in y"""
    import collections

    counter = collections.Counter(y)
    for i in range(len(counter)):
        if i not in counter.keys():
            counter[i] = 1

    num_samples = len(y)
    weights = [0] * len(counter)
    for idx in counter.keys():
        weights[idx] = 1.0 / (counter[idx] / num_samples)
    print("Inverse class weights: ")
    print(weights)

    return weights



import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
import pandas as pd
import re
import os
from scipy import stats
from sklearn.metrics import accuracy_score, f1_score, cohen_kappa_score, balanced_accuracy_score, confusion_matrix
import warnings
warnings.simplefilter("ignore", UserWarning)

from actinet.utils.utils import ACTIVITY_LABELS_DICT


class DivDict(dict):
    """Dictionary subclass that allows division by a number."""
    def __truediv__(self, n):
        if not isinstance(n, (int, float)):  
            raise TypeError("Can only divide by a number (int or float)")
        return DivDict({k: v / n for k, v in self.items()})


def calculate_metrics(y_true, y_pred):
    """Calculates accuracy, F1, Cohen's Kappa, and balanced accuracy."""
    accuracy = accuracy_score(y_true, y_pred)
    f1 = f1_score(y_true, y_pred, average='macro')
    kappa = cohen_kappa_score(y_true, y_pred)
    bacc = balanced_accuracy_score(y_true, y_pred)
    return accuracy, f1, kappa, bacc


def plot_confusion_matrix(y_true, y_pred, activity_labels, title="", ax=None, figsize=(8, 8), fontsize=20):
    """Plots a confusion matrix with heatmap annotations."""


    cm = confusion_matrix(y_true, y_pred, labels=activity_labels)
    cm_norm = cm.astype('float') / cm.sum(axis=1, keepdims=True)
    
    annotations = np.array([[f'{p:.3f}\n({n:,})' for p, n in zip(row_norm, row)] for row_norm, row in zip(cm_norm, cm)])
    
    if ax is None:
        _, ax = plt.subplots(figsize=figsize, dpi=800)

    sns.heatmap(cm_norm, annot=annotations, fmt='', cmap="Blues", cbar=False, 
                annot_kws={"size": fontsize*0.9}, ax=ax)
    ax.set_title(title, fontsize=fontsize)
    ax.set_xlabel('Predicted Label')
    ax.set_ylabel('True Label')

    ax.set_xticks(np.arange(len(activity_labels)) + 0.5)
    ax.set_yticks(np.arange(len(activity_labels)) + 0.5)
    ax.set_xticklabels(activity_labels, fontsize=fontsize*0.8)
    ax.set_yticklabels(activity_labels, fontsize=fontsize*0.8)


def build_confusion_matrix_data(results: pd.DataFrame, age_band=None, sex=None):
    """Extracts ground truth and predicted labels based on filtering conditions."""
    model_results = results.copy()
    if age_band is not None:
        model_results = model_results[model_results['Age Band'] == age_band]
    if sex is not None:
        model_results = model_results[model_results["Sex"] == sex]

    y_true_bbaa = np.hstack(model_results.loc[model_results["Model"]=="accelerometer", 'True'])
    y_pred_bbaa = np.hstack(model_results.loc[model_results["Model"]=="accelerometer", 'Predicted'])
    y_true_actinet = np.hstack(model_results.loc[model_results["Model"]=="actinet", 'True'])
    y_pred_actinet = np.hstack(model_results.loc[model_results["Model"]=="actinet", 'Predicted'])

    population = len(model_results['Participant'].unique())
    
    return y_true_bbaa, y_pred_bbaa, y_true_actinet, y_pred_actinet, population


def plot_and_save_fig(fig, save_path=None):
    """Displays and optionally saves the figure as a PDF."""
    plt.show()
    if save_path:
        os.makedirs(os.path.dirname(save_path), exist_ok=True)
        fig.savefig(save_path, format='pdf', dpi=800, bbox_inches='tight')


def generate_confusion_matrices(results_df, activity_labels, group_by=None, save_path=None, fontsize=20):
    """Generates and plots confusion matrices for different subgroups.""" 
    if group_by is None:  # Full population
        fig, axs = plt.subplots(1, 2, sharey=True, figsize=(12, 6), dpi=800)
        fig.suptitle("Confusion matrices for full Capture-24 population\nusing 5-fold group cross-validation", 
                     fontsize=fontsize)
        
        y_true_bbaa, y_pred_bbaa, y_true_actinet, y_pred_actinet, _ = build_confusion_matrix_data(results_df)
        
        plot_confusion_matrix(y_true_bbaa, y_pred_bbaa, activity_labels, 'Baseline', ax=axs[0], fontsize=fontsize*0.8)
        plot_confusion_matrix(y_true_actinet, y_pred_actinet, activity_labels, 'ActiNet', ax=axs[1], fontsize=fontsize*0.8)
        fig.tight_layout()
        plot_and_save_fig(fig, save_path)
    
    else:
        unique_groups = results_df[group_by].cat.categories
        fig = plt.figure(figsize=(12, 6*len(unique_groups)), dpi=800, constrained_layout=True)
        fig.suptitle(f"Confusion matrices for different {group_by.lower()} in " +
                     "Capture-24 population\nusing 5-fold group cross-validation", fontsize=fontsize)

        subfigs = fig.subfigures(nrows=len(unique_groups), ncols=1)
        
        for group, subfig in zip(unique_groups, subfigs):
            y_true_bbaa, y_pred_bbaa, y_true_actinet, y_pred_actinet, population =\
                build_confusion_matrix_data(results_df, **{group_by.replace(' ', '_').lower(): group})

            subfig.suptitle(f"{group_by}: {group} (n = {population})", fontsize=fontsize*0.9)
            axs = subfig.subplots(nrows=1, ncols=2, sharey=True)

            plot_confusion_matrix(y_true_bbaa, y_pred_bbaa, activity_labels, 'Baseline', ax=axs[0], fontsize=fontsize*0.8)
            plot_confusion_matrix(y_true_actinet, y_pred_actinet, activity_labels, 'ActiNet', ax=axs[1], fontsize=fontsize*0.8)

        plot_and_save_fig(fig, save_path)


def plot_model_performance(results, metric='Macro F1', modulus=0):
    """Plots a boxplot of model performance with optional participant-wise lines."""
    plt.figure(figsize=(10, 6), dpi=1000)
    with sns.color_palette("Set1"):
        sns.boxplot(x="Model", y=metric, hue="Model", data=results, width=0.3, showfliers=False)
    sns.stripplot(x="Model", y=metric, data=results, 
                  jitter=True, color='black', alpha=0.3)

    model1 = results[results['Model'] == results["Model"].unique()[0]][metric]
    model2 = results[results['Model'] == results["Model"].unique()[1]][metric]
    _, p_value = stats.ttest_rel(model1, model2)

    i = 0
    for pid in results['Participant'].unique():
        if modulus and i%modulus == 0:
            pid_df = results[results['Participant'] == pid]
            plt.plot(pid_df['Model'], pid_df[metric], marker='', 
                     linestyle='-', color='grey', alpha=0.3)
        i += 1

    plt.title(f"Comparison of model {metric}\np-value from paired t-test: {p_value:.2g}", fontsize=16)


def plot_difference_boxplots(df):
    """Plots boxplots of the differences in performance between actinet and accelerometer"""
    metrics = ['Accuracy', 'Macro F1', 'Cohen Kappa']
    differences = []
    
    for metric in metrics:
        model1 = df[df['Model'] == 'accelerometer'][metric].reset_index(drop=True)
        model2 = df[df['Model'] == 'actinet'][metric].reset_index(drop=True)
        difference = model2 - model1
        differences.append(pd.DataFrame({
            'Metric': metric,
            'Difference': difference
        }))
    
    diff_df = pd.concat(differences)
    
    plt.figure(figsize=(10, 6), dpi=800)
    sns.boxplot(x='Difference', y='Metric', hue='Metric', data=diff_df, orient='h')
    
    for metric in metrics:
        metric_data = diff_df[diff_df['Metric'] == metric]['Difference']
        mean_diff = metric_data.mean()
        std_diff = metric_data.std()
        plt.text(x=metric_data.quantile(.25)-0.01, y=metrics.index(metric)-0.3, 
                 s=f'Mean: {mean_diff:.3f}\nStd: {std_diff:.3f}', 
                 ha='right', va='center', color='black', fontsize=10, weight='bold')

    plt.title('Difference between Actinet and Accelerometer Models for Each Metric')
    plt.xlabel('Performance difference (actinet - accelerometer)')
    plt.ylabel('Metric')
    plt.savefig("outputs/har_activity_performance_difference_boxplots.png", bbox_inches='tight', dpi=800)
    plt.show()


def plot_boxplots(df, x, y='Macro F1', hue='Model'):
    """Plots boxplots of model performance by a specified variable"""
    _, ax = plt.subplots(figsize=(5, 3), dpi=1000)
    with sns.color_palette("Set1"):
        sns.boxplot(data=df, x=x, y=y, hue=hue, ax=ax)
    ax.set_xlabel("Age Band")
    ax.set_ylabel(f"{y} Score")
    plt.title(f"{y} by {x}")
    plt.show()


def extract_activity_predictions(results: pd.DataFrame, activity, age_band=None, sex=None, return_true_labels=False):
    """Extracts incidence of predicted activity label for actinet and accelerometer based on filtering conditions."""    
    model_results = results.copy()
    if age_band is not None:
        model_results = model_results[model_results['Age Band'] == age_band]
    if sex is not None:
        model_results = model_results[model_results["Sex"] == sex]

    activity_bbaa_pred = np.array([x.get(activity, 0) for x in model_results.loc[model_results["Model"] == "accelerometer", "Pred_dict"]])
    activity_actinet_pred = np.array([x.get(activity, 0) for x in model_results.loc[model_results["Model"] == "actinet", "Pred_dict"]])

    population = len(model_results['Participant'].unique())

    if return_true_labels:
        activity_bbaa_true = np.array([x.get(activity, 0) for x in model_results.loc[model_results["Model"] == "accelerometer",
                                                                                     "True_dict"]])
        activity_actinet_true = np.array([x.get(activity, 0) for x in model_results.loc[model_results["Model"] == "actinet", 
                                                                                        "True_dict"]])
        return activity_bbaa_pred, activity_actinet_pred, activity_bbaa_true, activity_actinet_true, population
    
    else:
        return activity_bbaa_pred, activity_actinet_pred, population


def bland_altman_plot(col1, col2, plot_label: str, anno_label: str, output_dir='',
                      col1_label='Baseline', col2_label='ActiNet',
                      display_plot=False, show_y_label=False, ax=None, fontsize=20):
    """Generates a Bland-Altman plot for two columns of data."""
    dat = pd.DataFrame({'col1': col1, 'col2': col2})
    pearson_cor = dat.corr().iloc[0, 1]
    diffs = dat['col2'] - dat['col1'] 
    mean_diff = np.mean(diffs)
    sd_diff = np.std(diffs, ddof=1)
    lower_loa = mean_diff - 1.96 * sd_diff
    upper_loa = mean_diff + 1.96 * sd_diff
    
    mean_vals = (dat['col1'] + dat['col2']) / 2

    if ax is None:
        _, ax = plt.subplots(figsize=(10, 10), dpi=800)

    ax.scatter(mean_vals, diffs, color='black', alpha=1)
    ax.axhline(mean_diff, color='red', linestyle='-')
    ax.axhline(lower_loa, color='blue', linestyle='--')
    ax.axhline(upper_loa, color='blue', linestyle='--')
    ax.text(0.8 * max(mean_vals), mean_diff, f'Mean Diff = {mean_diff:.2f}', va='bottom', color='red', fontsize=fontsize*0.8)
    ax.text(0.8 * max(mean_vals), lower_loa, f'Lower LoA = {lower_loa:.2f}', va='bottom', color='blue', fontsize=fontsize*0.8)
    ax.text(0.8 * max(mean_vals), upper_loa, f'Upper LoA = {upper_loa:.2f}', va='bottom', color='blue', fontsize=fontsize*0.8)

    ax.set_title(f'{ACTIVITY_LABELS_DICT[anno_label][plot_label]} [hours]\nPearson correlation: {pearson_cor:.3f}', fontsize=fontsize)

    ax.set_xlabel(f'({col2_label} + {col1_label}) / 2', fontsize=fontsize*0.9)
    
    if show_y_label:
        ax.set_ylabel(f'{col2_label} - {col1_label}', fontsize=fontsize*0.9)
    
    ax.tick_params(axis='both', which='both', labelsize=14)

    if display_plot and ax is None:
        plt.show()
    
    if ax is None:
        os.makedirs(output_dir, exist_ok=True)
        plot_path = os.path.join(output_dir, f'ba_{plot_label}_{col1_label}_vs_{col2_label}.png')
        plt.savefig(plot_path, bbox_inches='tight')
        plt.close()


def generate_bland_altman_plots(results_df, activities, anno_label, group_by=None, save_path=None, 
                                fontsize=20, compare_to_true=False, subset=""):
    """Generates Bland-Altman plots for different activities, optionally stratified by a subgroup."""
    if group_by is None:  # Full population
        fig, axs = plt.subplots(1, 4, figsize=(15, 6), dpi=800, sharey=True)      
        axs = axs.flatten()

        for i, activity in enumerate(activities):
            activity_bbaa_pred, activity_actinet_pred, activity_bbaa_true, activity_actinet_true, population =\
                 extract_activity_predictions(results_df, activity, return_true_labels=True)
            
            if compare_to_true=='bbaa':
                bland_altman_plot(activity_bbaa_true, activity_bbaa_pred, activity, anno_label,
                                  ax=axs[i], show_y_label=i==0, fontsize=fontsize*0.8,
                                  col1_label='Ground Truth', col2_label='Baseline')
            elif compare_to_true=='actinet':
                bland_altman_plot(activity_actinet_true, activity_actinet_pred, activity, anno_label,
                                  ax=axs[i], show_y_label=i==0, fontsize=fontsize*0.8,
                                  col1_label='Ground Truth', col2_label='ActiNet')
            elif compare_to_true is False:
                bland_altman_plot(activity_bbaa_pred, activity_actinet_pred, activity, anno_label,
                                  ax=axs[i], show_y_label=i==0, fontsize=fontsize*0.8,
                                  col1_label='Baseline', col2_label='ActiNet')
            else:
                raise ValueError("compare_to_true must be either False, 'bbaa' or 'actinet'")

        subset_in_title = f" {subset} " if subset else " "
        fig.suptitle(f"Bland-Altman plots for Capture-24{subset_in_title}population (n={population})", 
                     fontsize=fontsize)

    else:
        unique_groups = results_df[group_by].cat.categories
        fig = plt.figure(figsize=(15, 6*len(unique_groups)), dpi=800, constrained_layout=True)
        subfigs = fig.subfigures(nrows=len(unique_groups), ncols=1)

        for subfig, group in zip(subfigs, unique_groups):
            axs = subfig.subplots(nrows=1, ncols=len(activities), sharex=False, sharey=True)

            for i, activity in enumerate(activities):
                activity_bbaa_pred, activity_actinet_pred, activity_bbaa_true, activity_actinet_true, population =\
                    extract_activity_predictions(results_df, activity, **{group_by.replace(' ', '_').lower(): group}, 
                                                 return_true_labels=True)
                if compare_to_true=='bbaa':
                    bland_altman_plot(activity_bbaa_true, activity_bbaa_pred, activity, anno_label,
                                      ax=axs[i], show_y_label=i==0, fontsize=fontsize*0.8,
                                      col1_label='Ground Truth', col2_label='Baseline')
                elif compare_to_true=='actinet':
                    bland_altman_plot(activity_actinet_true, activity_actinet_pred, activity, anno_label,
                                      ax=axs[i], show_y_label=i==0, fontsize=fontsize*0.8,
                                      col1_label='Ground Truth', col2_label='ActiNet')
                elif compare_to_true is False:
                    bland_altman_plot(activity_bbaa_pred, activity_actinet_pred, activity, anno_label,
                                      ax=axs[i], show_y_label=i==0, fontsize=fontsize*0.8,
                                      col1_label='Baseline', col2_label='ActiNet')
                else:
                    raise ValueError("compare_to_true must be either False, 'bbaa' or 'actinet'")

            subfig.suptitle(f"{group_by}: {group} (n={population})", fontsize=fontsize*0.9)

        fig.suptitle(f"Bland-Altman plots for Capture-24 population by {group_by}", fontsize=fontsize)

    fig.tight_layout()
    plot_and_save_fig(fig, save_path=save_path)
            
    plt.close(fig)


def build_mae_cell(true_values, pred_values):
    """Builds a MAE cell for a given set of true and predicted values."""
    mae = np.abs(true_values - pred_values)
    return f"{np.mean(mae):.3f} ± {np.std(mae):.3f}"


def build_pvalue_cell(true_values, pred_values):
    """Builds a p-value cell for a given set of true and predicted values."""
    _, p_value = stats.ttest_rel(true_values, pred_values)
    if p_value < 0.001:
        return "<0.001"
    return f"{p_value:.3f}"


def build_mae_table(df: pd.DataFrame, activities):
    df_maes = pd.DataFrame(columns=activities, index=['Baseline', 'ActiNet', 'p-value'], dtype=float)
    for activity in activities:
        activity_bbaa_pred, activity_actinet_pred, activity_bbaa_true, activity_actinet_true, _ =\
            extract_activity_predictions(df, activity, return_true_labels=True)
        df_maes.loc['Baseline', activity] = build_mae_cell(activity_bbaa_true, activity_bbaa_pred)
        df_maes.loc['ActiNet', activity] = build_mae_cell(activity_actinet_true, activity_actinet_pred)
        df_maes.loc['p-value', activity] = build_pvalue_cell(activity_bbaa_true, activity_bbaa_pred)
    return df_maes


def plot_errors(df: pd.DataFrame, activities, anno_label, group_by=None, save_path=None, fontsize=12):
    if group_by is None:
        all_errors = []

        for activity in activities:
            activity_bbaa_pred, activity_actinet_pred, activity_bbaa_true, activity_actinet_true, _ = \
                extract_activity_predictions(df, activity, return_true_labels=True)
            
            bbaa_errors = activity_bbaa_pred - activity_bbaa_true
            actinet_errors = activity_actinet_pred - activity_actinet_true

            for err in bbaa_errors:
                all_errors.append({'Activity': ACTIVITY_LABELS_DICT[anno_label][activity],
                                   'Error': err, 'Model': 'Baseline'})
            for err in actinet_errors:
                all_errors.append({'Activity': ACTIVITY_LABELS_DICT[anno_label][activity],
                                   'Error': err, 'Model': 'ActiNet'})

        error_df = pd.DataFrame(all_errors)

        fig, ax = plt.subplots(figsize=(2*len(activities), 4), dpi=1000)
        with sns.color_palette("Set1"):
            sns.boxplot(x="Activity", y="Error", hue="Model", data=error_df,
                        width=0.5, showfliers=False, ax=ax)

        ax.set_xlabel("Activity", fontsize=fontsize)
        ax.set_ylabel("Error in total estimated activity [hours]\n(model - ground-truth)", fontsize=fontsize)
        ax.axhline(0, color='red', linestyle='--', linewidth=1)
        ax.grid(axis='y', alpha=0.7)

        handles, labels = ax.get_legend_handles_labels()
        n = len(set(error_df['Model']))
        ax.legend(handles[:n], labels[:n], fontsize=fontsize*0.8)
        plt.title("Box plots for the error distribution of activity recognition models' outputs\n" +\
                  "evaluated on the full Capture-24 population", fontsize=fontsize)
        fig.tight_layout()

    else:
        unique_groups = df[group_by].cat.categories
        fig, axs = plt.subplots(nrows=len(unique_groups), ncols=1,
                                figsize=(2*len(activities), 4*len(unique_groups)),
                                dpi=1000, constrained_layout=True)
        
        if len(unique_groups) == 1:
            axs = [axs]

        for ax, group in zip(axs, unique_groups):
            sub_df = df[df[group_by] == group]
            all_errors = []

            for activity in activities:
                activity_bbaa_pred, activity_actinet_pred, activity_bbaa_true, activity_actinet_true, _ = \
                    extract_activity_predictions(sub_df, activity, return_true_labels=True)

                bbaa_errors = activity_bbaa_pred - activity_bbaa_true
                actinet_errors = activity_actinet_pred - activity_actinet_true

                for err in bbaa_errors:
                    all_errors.append({'Activity': ACTIVITY_LABELS_DICT[anno_label][activity],
                                       'Error': err, 'Model': 'Baseline'})
                for err in actinet_errors:
                    all_errors.append({'Activity': ACTIVITY_LABELS_DICT[anno_label][activity],
                                       'Error': err, 'Model': 'ActiNet'})

            error_df = pd.DataFrame(all_errors)

            with sns.color_palette("Set1"):
                sns.boxplot(x="Activity", y="Error", hue="Model", data=error_df,
                            width=0.5, showfliers=False, ax=ax)

            ax.set_xlabel("Activity", fontsize=fontsize)
            ax.set_ylabel("Error in total estimated activity [hours]\n(model - ground-truth)", fontsize=fontsize)
            ax.axhline(0, color='red', linestyle='--', linewidth=1)
            ax.grid(axis='y', alpha=0.7)
            ax.set_title(f"{group_by}: {group}", fontsize=fontsize)

            handles, labels = ax.get_legend_handles_labels()
            n = len(set(error_df['Model']))
            ax.legend(handles[:n], labels[:n], fontsize=fontsize*0.8)

        plt.suptitle("Box plots for the error distribution of activity recognition models' outputs\n" +\
                     f"evaluated on the Capture-24 population by {group_by}", fontsize=fontsize)
    if save_path:
        fig.savefig(save_path, bbox_inches='tight')
    plt.show()


def convert_version(s: str) -> str:
    """ Converts a version string from 'x.y.z+number' to 'vX-Y-Z'. """
    match = re.match(r"(\d+)\.(\d+)\.(\d+)(?:\+(\d+))?", s)
    if not match:
        raise ValueError(f"String '{s}' is not in the expected format")
    
    major, minor, patch, _ = match.groups()
    base = f"v{major}-{minor}-{patch}"
    return base



from sklearn.model_selection import StratifiedGroupKFold
from sklearn.preprocessing import LabelEncoder
from sklearn.metrics import (
    classification_report,
    accuracy_score,
    f1_score,
    cohen_kappa_score,
)
import numpy as np
import pandas as pd
import os

from actinet.models import ActivityClassifier, RFActivityClassifier
from actinet.hmm import HMM
from actinet.utils.utils import safe_indexer

WINSEC = 30


def evaluate_preprocessing(
    classifier: ActivityClassifier,
    X,
    Y,
    groups=None,
    T=None,
    weights_path="models/weights.pt",
    verbose=True,
):
    skf = StratifiedGroupKFold(n_splits=5)

    le = LabelEncoder().fit(Y)
    Y_encoded = le.transform(Y)

    Y_preds = np.empty_like(Y_encoded)

    for fold, (train_index, test_index) in enumerate(skf.split(X, Y_encoded, groups)):
        X_train, X_test = X[train_index], X[test_index]
        y_train, y_test = Y_encoded[train_index], Y_encoded[test_index]
        groups_train = safe_indexer(groups, train_index)
        t_train = safe_indexer(T, train_index)

        classifier.fit(
            X_train,
            y_train,
            groups_train,
            t_train,
            weights_path.format(fold),
            n_splits=1,
        )
        y_pred = classifier.predict(X_test, False)

        if verbose:
            print(
                f"Fold {fold+1} Test Scores - Accuracy: {accuracy_score(y_test, y_pred):.3f}, "
                + f"Macro F1: {f1_score(y_test, y_pred, average='macro'):.3f}"
            )

        Y_preds[test_index] = y_pred

    Y_preds = le.inverse_transform(Y_preds)

    if verbose:
        print(classification_report(Y, Y_preds))

    return Y_preds


def evaluate_models(
    actinet_classifier: ActivityClassifier,
    rf_classifier: RFActivityClassifier,
    X_actinet,
    X_rf,
    Y_actinet,
    Y_rf,
    groups_actinet,
    groups_rf,
    T_actinet=None,
    T_rf=None,
    sleep_tol='1H',
    remove_naps=False,
    weights_path="models/weights.pt",
    out_dir=None,
    verbose=True,
):
    skf = StratifiedGroupKFold(n_splits=5)

    le = LabelEncoder().fit(Y_rf)
    Y_encoded_rf = le.transform(Y_rf)
    Y_encoded_actinet = le.transform(Y_actinet)

    Y_preds_rf = np.empty_like(Y_encoded_rf)
    Y_preds_actinet = np.empty_like(Y_encoded_actinet)

    results_rf = []
    results_actinet = []

    for fold, (train_index, test_index) in enumerate(
        skf.split(X_rf, Y_encoded_rf, groups_rf)
    ):
        if verbose:
            print(f"======== Evalating Fold {fold+1} ========")
        # Ensure the same train and test split for groups are used in both models in each fold
        train_index_actinet = np.isin(groups_actinet, np.unique(groups_rf[train_index]))
        test_index_actinet = np.isin(groups_actinet, np.unique(groups_rf[test_index]))

        train_index_rf = np.isin(groups_rf, np.unique(groups_rf[train_index]))
        test_index_rf = np.isin(groups_rf, np.unique(groups_rf[test_index]))

        # Train test split for actinet model
        X_train_actinet, X_test_actinet = (
            X_actinet[train_index_actinet],
            X_actinet[test_index_actinet],
        )
        y_train_actinet, y_test_actinet = (
            Y_encoded_actinet[train_index_actinet],
            Y_encoded_actinet[test_index_actinet],
        )
        groups_train_actinet = groups_actinet[train_index_actinet]
        groups_test_actinet = groups_actinet[test_index_actinet]

        t_train_actinet = safe_indexer(T_actinet, train_index_actinet)
        t_test_actinet = safe_indexer(T_actinet, test_index_actinet)

        # Train test split for accelerometer model
        X_train_rf, X_test_rf = X_rf[train_index_rf], X_rf[test_index_rf]
        y_train_rf, y_test_rf = (
            Y_encoded_rf[train_index_rf],
            Y_encoded_rf[test_index_rf],
        )
        t_train_rf = safe_indexer(T_rf, train_index_rf)
        t_test_rf = safe_indexer(T_rf, test_index_rf)

        groups_train_rf = groups_rf[train_index_rf]
        groups_test_rf = groups_rf[test_index_rf]

        actinet_classifier.fit(
            X_train_actinet,
            y_train_actinet,
            groups_train_actinet,
            t_train_actinet,
            weights_path.format(fold),
            n_splits=1,
        )

        y_pred_actinet = actinet_classifier.predict(
            X_test_actinet, t_test_actinet, True, sleep_tol, remove_naps
        ).astype(int)

        rf_classifier.fit(
            X_train_rf,
            y_train_rf,
            groups_train_rf,
            t_train_rf
        )

        y_pred_rf = rf_classifier.predict(X_test_rf, t_test_rf, True, sleep_tol, remove_naps)

        # Display model performance for each fold
        if verbose:
            print(
                f"Actinet test scores for fold {fold+1}\n"
                + f"Accuracy: {accuracy_score(y_test_actinet, y_pred_actinet):.3f}, "
                + f"Macro F1: {f1_score(y_test_actinet, y_pred_actinet, average='macro'):.3f}, "
                + f"Kappa: {cohen_kappa_score(y_test_actinet, y_pred_actinet):.3f}"
            )
            print(
                f"Accelerometer test scores for fold {fold+1}\n"
                + f"Accuracy: {accuracy_score(y_test_rf, y_pred_rf):.3f}, "
                + f"Macro F1: {f1_score(y_test_rf, y_pred_rf, average='macro'):.3f}, "
                + f"Kappa: {cohen_kappa_score(y_test_rf, y_pred_rf):.3f}"
            )

        Y_preds_actinet[test_index_actinet] = y_pred_actinet
        Y_preds_rf[test_index_rf] = y_pred_rf

        results_actinet.append(
            {
                "fold": [fold] * len(y_pred_actinet),
                "group": groups_test_actinet,
                "Y_pred": le.inverse_transform(y_pred_actinet),
                "Y_true": le.inverse_transform(y_test_actinet),
            }
        )
        results_rf.append(
            {
                "fold": [fold] * len(y_pred_rf),
                "group": groups_test_rf,
                "Y_pred": le.inverse_transform(y_pred_rf),
                "Y_true": le.inverse_transform(y_test_rf),
            }
        )

    Y_preds_actinet = le.inverse_transform(Y_preds_actinet)
    Y_preds_rf = le.inverse_transform(Y_preds_rf)

    # Report performance across all folds
    if verbose:
        print("Actinet performance:")
        print(classification_report(Y_actinet, Y_preds_actinet))
        print("Accelerometer performance:")
        print(classification_report(Y_rf, Y_preds_rf))

    # Save results to pickle files
    results_actinet = pd.DataFrame(results_actinet)
    results_rf = pd.DataFrame(results_rf)

    if out_dir is not None:
        os.makedirs(out_dir, exist_ok=True)
        results_actinet.to_pickle(f"{out_dir}/actinet_results.pkl")
        results_rf.to_pickle(f"{out_dir}/rf_results.pkl")

    return results_actinet, results_rf



import sys
import os.path

# https://github.com/python-versioneer/python-versioneer/issues/193
sys.path.insert(0, os.path.dirname(__file__))

from setuptools import setup, find_packages
import codecs

import versioneer


def main():

    with open("README.md", "r") as fh:
        long_description = fh.read()

    setup(
        name="actinet",
        version=versioneer.get_version(),  # Do not edit
        cmdclass=versioneer.get_cmdclass(),  # Do not edit
        python_requires=">=3.8, <4",
        description="Activity detection algorithm compatible with the UK Biobank Accelerometer Dataset",
        long_description=long_description,  # Do not edit. See README.md
        long_description_content_type="text/markdown",
        keywords="example, setuptools, versioneer",
        url="https://github.com/OxWearables/actinet",
        download_url="https://github.com/OxWearables/actinet",
        author=get_string("__author__"),  # Do not edit. see src/actinet/__init__.py
        maintainer=get_string(
            "__maintainer__"
        ),  # Do not edit. see src/actinet/__init__.py
        maintainer_email=get_string(
            "__maintainer_email__"
        ),  # Do not edit. See src/actinet/__init__.py
        license=get_string("__license__"),  # Do not edit. See src/actinet/__init__.py
        # This is for PyPI to categorize your project. See: https://pypi.org/classifiers/
        classifiers=[
            "Programming Language :: Python :: 3",
            "Topic :: Scientific/Engineering",
            "Topic :: Scientific/Engineering :: Bio-Informatics",
            "Topic :: Scientific/Engineering :: Information Analysis",
            "Topic :: Scientific/Engineering :: Medical Science Apps.",
        ],
        # Where to find the source code
        packages=find_packages(where="src", exclude=("test", "tests")),
        package_dir={"": "src"},
        # What other files to include, e.g. *.class if the package uses some Java code.
        package_data={"actinet": ["*.txt", "*.rst", "*.md"]},
        include_package_data=False,
        # Dependencies
        install_requires=[
            "actipy>=3.7.0",
            "numpy==1.24.*",
            "scipy==1.10.*",
            "pandas==2.0.*",
            "tqdm==4.64.*",
            "matplotlib==3.5.*",
            "joblib==1.2.*",
            "scikit-learn==1.1.1",
            "imbalanced-learn==0.9.1",
            "torch==1.13.*",
            "torchvision==0.14.*",
            "transforms3d==0.4.*",
        ],
        extras_require={
            "dev": [
                "versioneer",
                "twine",
                "ipython",
                "ipdb",
                "flake8",
                "autopep8",
                "tomli",
                "jupyter",
            ],
            "docs": [
                "sphinx>=4.2",
                "sphinx_rtd_theme>=1.0",
                "readthedocs-sphinx-search>=0.1",
                "docutils<0.18",
            ],
        },
        # Define entry points for command-line scripts, e.g.: `$ hello --name Alice`
        entry_points={
            "console_scripts": [
                "actinet=actinet.actinet:main",
                "actinet-collate-outputs=actinet.utils.collate_outputs:main",
                "accPlot=actinet.accPlot:main",
            ],
        },
    )


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def get_string(string, rel_path="src/actinet/__init__.py"):
    for line in read(rel_path).splitlines():
        if line.startswith(string):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError(f"Unable to find {string}.")


if __name__ == "__main__":
    main()



import json
import numpy as np
import pandas as pd
import os
from tqdm.auto import tqdm
from joblib import Parallel, delayed
import warnings
import actipy
from glob import glob
import subprocess

from actinet.utils.utils import is_good_window, resize

ACC_COLS = ["x", "y", "z"]


def load_data(
    datafile,
    sample_rate=100,
    annot_type=str,
    lowpass_hz=None,
    resample_rate=None,
):
    """
    Load data from a file and process it using actipy.

    Args:
        datafile (str): Path to the data file.
        sample_rate (int): Sampling rate of the data. Defaults to 100.
        annot_type (type): Type of annotation. Defaults to <class 'str'>.

    Returns:
        pd.DataFrame: Processed data.

    """
    if ".parquet" in datafile:
        data = pd.read_parquet(datafile)
        data.dropna(inplace=True)

    else:
        data = pd.read_csv(
            datafile,
            index_col=[0],
            parse_dates=True,
            dtype={"x": "f4", "y": "f4", "z": "f4", "annotation": annot_type},
        )

    data, _ = actipy.process(
        data,
        sample_rate,
        verbose=False,
        lowpass_hz=lowpass_hz,
        resample_hz=resample_rate,
    )
    return data


def make_windows(
    data,
    anno_dict,
    anno_label,
    winsec=30,
    sample_rate=100,
    resample_rate=30,
):
    """
    Create windows from the input data.

    Args:
        data (pd.DataFrame): Input data.
        anno_dict (pd.DataFrame): Annotation dictionary.
        anno_label (string): Column in annotation dictionary to convert raw annotations to activities.
        winsec (int): Window size in seconds.
        sample_rate (int): Sampling rate of the data.
        resample_rate (int): Resampling rate. Needs to be 30Hz for SSL model.

    Returns:
        tuple: Tuple containing the windowed data, labels, and timestamps.

    """
    X, Y, T = [], [], []

    for t, w in data.resample(f"{winsec}s", origin="start"):

        if len(w) < 1:
            continue

        t = t.to_numpy()

        x = w[ACC_COLS]

        annot = w["annotation"]

        if pd.isna(annot).all():  # skip if annotation is NA
            continue

        if not is_good_window(x, sample_rate * winsec, ACC_COLS):  # skip if bad window
            continue

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="Unable to sort modes")
            y = (
                anno_dict.loc[annot.dropna(), f"label:{anno_label}"]
                .mode(dropna=False)
                .iloc[0]
            )

        X.append(x.to_numpy())
        Y.append(y)
        T.append(t)

    X = np.stack(X)
    Y = np.stack(Y)
    T = np.stack(T)

    if resample_rate != sample_rate:
        X = resize(X, int(resample_rate * winsec))

    return X, Y, T


def load_all_and_make_windows(
    datafiles,
    annofile,
    out_dir=None,
    anno_label="Walmsley2020",
    sample_rate=100,
    winsec=30,
    resample_rate=30,
    lowpass_hz=None,
    downsampling_method="nn",
    n_jobs=1,
):
    """
    Load data from multiple files, create windows, and save the results.

    Args:
        datafiles (list): List of data file paths.
        annofile (str): Path to the annotation file.
        out_dir (str | None): Output directory to save the results. Outputs not saved if None.
        n_jobs (int): Number of parallel jobs.

    Returns:
        tuple: Tuple containing the windowed data, labels, timestamps, and participant IDs.

    """

    def worker(datafile):
        if downsampling_method == "nn":
            X, Y, T = make_windows(
                load_data(
                    datafile,
                    sample_rate,
                    resample_rate=resample_rate,
                    lowpass_hz=lowpass_hz,
                ),
                anno_dict,
                anno_label,
                winsec,
                resample_rate,
                resample_rate,
            )
        elif downsampling_method == "linear":
            X, Y, T = make_windows(
                load_data(
                    datafile,
                    sample_rate,
                    lowpass_hz=lowpass_hz,
                ),
                anno_dict,
                anno_label,
                winsec,
                sample_rate,
                resample_rate,
            )
        else:
            raise ValueError("Invalid downsampling method")

        pid = os.path.basename(datafile).split(".")[
            0
        ]  # participant ID based on file name
        pid = np.asarray([pid] * len(X))
        return X, Y, T, pid

    anno_dict = pd.read_csv(annofile, index_col="annotation", dtype="string")

    X, Y, T, P = zip(
        *Parallel(n_jobs=n_jobs)(
            delayed(worker)(datafile)
            for datafile in tqdm(datafiles, desc="Load and making windows: ")
        )
    )

    X = np.vstack(X)
    Y = np.hstack(Y)
    T = np.hstack(T)
    P = np.hstack(P)

    if out_dir:
        info = {
            "sample_rate": sample_rate,
            "winsec": winsec,
            "resample_rate": resample_rate,
            "lowpass_hz": lowpass_hz,
            "downsampling_method": downsampling_method,
            "anno_label": anno_label,
        }

        # Save arrays for future use
        os.makedirs(out_dir, exist_ok=True)
        np.save(f"{out_dir}/X.npy", X)
        np.save(f"{out_dir}/Y.npy", Y)
        np.save(f"{out_dir}/T.npy", T)
        np.save(f"{out_dir}/pid.npy", P)

        with open(f"{out_dir}/info.json", "w") as f:
            json.dump(info, f, indent=4)

    return X, Y, T, P


def make_labels(
    data,
    anno_dict,
    anno_label,
    sample_rate=100,
    winsec=30,
):
    Y, T = [], []
    for t, w in data.resample(f"{winsec}s", origin="start"):
        if len(w) < 1:
            continue

        t = t.to_numpy()

        annot = w["annotation"]

        if pd.isna(annot).all():  # skip if annotation is NA
            y = np.nan

        elif not is_good_window(
            w, sample_rate * winsec, ACC_COLS
        ):  # skip if bad window
            y = np.nan

        else:
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", message="Unable to sort modes")
                y = (
                    anno_dict.loc[annot.dropna(), f"label:{anno_label}"]
                    .mode(dropna=False)
                    .iloc[0]
                )

        Y.append(y)
        T.append(t)

    Y = np.stack(Y)
    T = np.stack(T)

    return Y, T


def extract_accelerometer_features(n_jobs):
    def process_file(file_number):
        filename = f"P{file_number:03}.csv.gz"

        if len(glob(f"data/capture24/bbaa/P{file_number:03}*")) != 4:
            command = f'accProcess data/capture24/{filename} --csvTimeFormat "yyyy-MM-dd HH:mm:ss.SSSSSS" --csvTimeXYZTempColsIndex "0,1,2,3" -o "data/capture24/bbaa" --activityClassification False --deleteIntermediateFiles False'
            process = subprocess.run(command, shell=True, capture_output=True)

            return process

    Parallel(n_jobs=n_jobs)(
        delayed(process_file)(file_number) for file_number in tqdm(range(1, 152))
    )


def prepare_participant_accelerometer_data(pid, annotations_file, anno_label, verbose=False):
    raw_file_name = f"data/capture24/P{pid:03}.csv.gz"
    features_file_name = f"data/capture24/bbaa/P{pid:03}-epoch.csv.gz"

    raw_file = pd.read_csv(
        raw_file_name,
        index_col="time",
        parse_dates=True,
        dtype={"x": "f4", "y": "f4", "z": "f4", "annotation": str},
    )
    features_file = pd.read_csv(features_file_name, index_col="time")

    if verbose:
        print(
            f"First accelerometer timestamp: {features_file.index[0]}\n"
            + f"First actipy timestamp: {raw_file.index[0]}\n"
            + f"Last accelerometer timestamp: {features_file.index[-1]}\n"
            + f"Last actipy timestamp: {raw_file.index[-1]}\n"
        )

    Y, T = make_labels(
        raw_file,
        pd.read_csv(annotations_file, index_col="annotation", dtype="string"),
        anno_label,
    )

    features = (
        ["enmoTrunc", "enmoAbs", "xMean", "yMean", "zMean", "xRange"]
        + ["yRange", "zRange", "xStd", "yStd", "zStd", "xyCov", "xzCov", "yzCov"]
        + ["mean", "sd", "coefvariation", "median", "min", "max", "25thp", "75thp"]
        + ["autocorr", "corrxy", "corrxz", "corryz", "avgroll", "avgpitch"]
        + ["avgyaw", "sdroll", "sdpitch", "sdyaw", "rollg", "pitchg", "yawg"]
        + ["fmax", "pmax", "fmaxband", "pmaxband", "entropy", "fft1", "fft2"]
        + ["fft3", "fft4", "fft5", "fft6", "fft7", "fft8", "fft9", "fft10", "MAD"]
        + ["MPD", "skew", "kurt", "avgArmAngel", "avgArmAngelAbsDiff", "f1", "p1"]
        + ["f2", "p2", "f625", "p625", "totalPower"]
    )

    X = features_file[features].to_numpy()
    Y, T = Y[: len(X)], T[: len(X)]
    P = np.array([f"P{pid:03}"] * len(X))

    mask = ~(pd.isna(Y) | (Y == "nan"))
    X, Y, T, P = X[mask], Y[mask], T[mask], P[mask]

    return X, Y, T, P


def prepare_accelerometer_data(annotation_file, anno_label, out_dir, n_jobs):
    X, Y, T, P = zip(
        *Parallel(n_jobs=n_jobs)(
            delayed(prepare_participant_accelerometer_data)(
                file_number, annotation_file, anno_label
            )
            for file_number in tqdm(range(1, 152))
        )
    )

    X = np.vstack(X)
    Y = np.hstack(Y)
    T = np.hstack(T)
    P = np.hstack(P)

    if out_dir:
        out_path = os.path.join(out_dir, f"prepared/{anno_label}")
        # Save arrays for future use
        os.makedirs(out_path, exist_ok=True)
        np.save(f"{out_path}/X.npy", X)
        np.save(f"{out_path}/Y.npy", Y)
        np.save(f"{out_path}/T.npy", T)
        np.save(f"{out_path}/pid.npy", P)

    return X, Y, T, P



name = "actinet"
__author__ = "Aidan Acquah, Shing Chan, Aiden Doherty"
__maintainer__ = "Shing Chan"
__maintainer_email__ = "shing.chan@ndph.ox.ac.uk"
__license__ = "See LICENSE file."

__classifiers__ = {
    'walmsley': {
        'version': 'ssl-ukb-c24-rw-30s-20260128',
        'md5': 'a829041ba84b18084b4b2897fb1b36a6'
    },
    'willetts': {
        'version': 'ssl-ukb-c24-mw-30s-20260128',
        'md5': '59a9492b50ee922c7a31f88888c5f14b'
    }
}

from . import _version

__version__ = _version.get_versions()["version"]
